#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""TensorBay dataset version control related classes."""

from typing import Any, Dict, Generator, Optional, Union

from tensorbay.client.lazy import PagingList
from tensorbay.client.requests import Client
from tensorbay.client.status import Status
from tensorbay.client.struct import Branch, Commit, Draft, Tag
from tensorbay.exception import ResourceNotExistError, StatusError


class VersionControlMixin:  # pylint: disable=too-many-public-methods
    """A mixin class supporting version control methods.

    Arguments:
        _dataset_id: Dataset ID.
        _client: The client to interact between local and TensorBay.
        _status: The version control status of the dataset.

    """

    _dataset_id: str
    _client: Client
    _status: Status

    def _get_basehead(
        self, base: Optional[Union[str, int]] = None, head: Optional[Union[str, int]] = None
    ) -> str:
        if head:
            head = f"draft-{head}" if isinstance(head, int) else f"commit-{head}"
        else:
            if self._status.is_draft:
                head = f"draft-{self._status.draft_number}"
            else:
                head = f"commit-{self._status.commit_id}"

        if base:
            base = f"draft-{base}" if isinstance(base, int) else f"commit-{base}"
            basehead = f"{base}...{head}"
        else:
            basehead = head

        return basehead

    def _commit(self, title: str, description: str, tag: Optional[str] = None) -> str:
        post_data: Dict[str, Any] = {"title": title}
        post_data.update(self._status.get_status_info())

        if description:
            post_data["description"] = description

        if tag:
            post_data["tag"] = tag

        response = self._client.open_api_do("POST", "commits", self._dataset_id, json=post_data)
        return response.json()["commitId"]  # type: ignore[no-any-return]

    def _generate_drafts(
        self, status: Optional[str], branch_name: Optional[str], offset: int = 0, limit: int = 128
    ) -> Generator[Draft, None, int]:
        params = {"offset": offset, "limit": limit, "status": status, "branchName": branch_name}
        response = self._client.open_api_do("GET", "drafts", self._dataset_id, params=params).json()

        for item in response["drafts"]:
            yield Draft.loads(item)

        return response["totalCount"]  # type: ignore[no-any-return]

    def _close_draft(self, number: int) -> None:
        patch_data = {"status": "CLOSED"}
        self._client.open_api_do("PATCH", f"drafts/{number}", self._dataset_id, json=patch_data)

    def _generate_commits(
        self, revision: str, offset: int = 0, limit: int = 128
    ) -> Generator[Commit, None, int]:
        params: Dict[str, Any] = {"offset": offset, "limit": limit, "commit": revision}

        response = self._client.open_api_do(
            "GET", "commits", self._dataset_id, params=params
        ).json()

        for item in response["commits"]:
            yield Commit.loads(item)

        return response["totalCount"]  # type: ignore[no-any-return]

    def _generate_branches(
        self, name: Optional[str] = None, offset: int = 0, limit: int = 128
    ) -> Generator[Branch, None, int]:
        params: Dict[str, Any] = {"offset": offset, "limit": limit}
        if name:
            params["name"] = name

        response = self._client.open_api_do(
            "GET", "branches", self._dataset_id, params=params
        ).json()

        for item in response["branches"]:
            yield Branch.loads(item)

        return response["totalCount"]  # type: ignore[no-any-return]

    def _generate_tags(
        self, name: Optional[str] = None, offset: int = 0, limit: int = 128
    ) -> Generator[Tag, None, int]:
        params: Dict[str, Any] = {"offset": offset, "limit": limit}
        if name:
            params["name"] = name

        response = self._client.open_api_do("GET", "tags", self._dataset_id, params=params).json()

        for item in response["tags"]:
            yield Tag.loads(item)

        return response["totalCount"]  # type: ignore[no-any-return]

    def _delete_branch(self, name: str) -> None:
        """Delete a branch without checking current branch.

        Arguments:
            name: The name of the branch to be deleted.

        """
        delete_data: Dict[str, Any] = {"name": name}

        self._client.open_api_do("DELETE", "branches", self._dataset_id, json=delete_data)

    def checkout(self, revision: Optional[str] = None, draft_number: Optional[int] = None) -> None:
        """Checkout to commit or draft.

        Arguments:
            revision: The information to locate the specific commit, which can be the commit id,
                the branch, or the tag.
            draft_number: The draft number.

        Raises:
            TypeError: When both commit and draft number are provided or neither.

        """
        if revision is None and draft_number is None:
            raise TypeError("Neither revision nor draft number is given, please give one")
        if revision is not None and draft_number is not None:
            raise TypeError("Both revision and draft number are given, please only give one")

        if revision:
            try:
                branch = self.get_branch(revision)
                self._status.checkout(commit_id=branch.commit_id)
                self._status.branch_name = branch.name
            except ResourceNotExistError:
                self._status.checkout(commit_id=self.get_commit(revision).commit_id)
                self._status.branch_name = None

        if draft_number:
            draft = self.get_draft(draft_number)
            self._status.checkout(draft_number=draft.number)
            self._status.branch_name = draft.branch_name

    def commit(self, title: str, description: str = "", *, tag: Optional[str] = None) -> None:
        """Commit the draft.

        Commit the draft based on the draft number stored in the dataset client.
        Then the dataset client will change the status to "commit"
        and store the branch name and commit id.

        Arguments:
            title: The commit title.
            description: The commit description.
            tag: A tag for current commit.

        """
        self._status.check_authority_for_draft()
        self._status.checkout(commit_id=self._commit(title, description, tag))

    def create_draft(
        self, title: str, description: str = "", branch_name: Optional[str] = None
    ) -> int:
        """Create a draft.

        Create a draft with the branch name. If the branch name is not given,
        create a draft based on the branch name stored in the dataset client.
        Then the dataset client will change the status to "draft"
        and store the branch name and draft number.

        Arguments:
            title: The draft title.
            description: The draft description.
            branch_name: The branch name.

        Returns:
            The draft number of the created draft.

        Raises:
            StatusError: When creating the draft without basing on a branch.

        """
        if not branch_name:
            branch_name = self._status.branch_name
            if not branch_name:
                raise StatusError(
                    message="Creating the draft without basing on a branch is not allowed"
                )
            self._status.check_authority_for_commit()

        post_data: Dict[str, Any] = {"branchName": branch_name, "title": title}

        if description:
            post_data["description"] = description

        response = self._client.open_api_do("POST", "drafts", self._dataset_id, json=post_data)
        draft_number: int = response.json()["draftNumber"]

        self._status.checkout(draft_number=draft_number)
        self._status.branch_name = branch_name
        return draft_number

    def get_draft(self, draft_number: Optional[int] = None) -> Draft:
        """Get the certain draft with the given draft number.

        Get the certain draft with the given draft number. If the draft number is not given,
        get the draft based on the draft number stored in the dataset client.

        Arguments:
            draft_number: The required draft number.
                If is not given, get the current draft.

        Returns:
            The :class:`.Draft` instance with the given number.

        Raises:
            TypeError: When the given draft number is illegal.
            ResourceNotExistError: When the required draft does not exist.

        """
        if draft_number is None:
            self._status.check_authority_for_draft()
            draft_number = self._status.draft_number

        if not draft_number:
            raise TypeError("The given draft number is illegal")

        for draft in self.list_drafts():
            if draft_number == draft.number:
                return draft

        raise ResourceNotExistError(resource="draft", identification=draft_number)

    def list_drafts(
        self, status: Optional[str] = "OPEN", branch_name: Optional[str] = None
    ) -> PagingList[Draft]:
        """List all the drafts.

        Arguments:
            status: The draft status which includes "OPEN", "CLOSED", "COMMITTED", "ALL" and None.
                    where None means listing open drafts.
            branch_name: The branch name.

        Returns:
            The PagingList of :class:`drafts<.Draft>`.

        """
        return PagingList(
            lambda offset, limit: self._generate_drafts(status, branch_name, offset, limit),
            128,
        )

    def update_draft(
        self,
        draft_number: Optional[int] = None,
        *,
        title: Optional[str] = None,
        description: Optional[str] = None,
    ) -> None:
        """Update the draft.

        Arguments:
            draft_number: The updated draft number.
                If is not given, update the current draft.
            title: The title of the draft.
            description: The description of the draft.

        """
        if draft_number is None:
            self._status.check_authority_for_draft()
            draft_number = self._status.draft_number

        patch_data: Dict[str, Any] = {}
        if title is not None:
            patch_data["title"] = title

        if description is not None:
            patch_data["description"] = description

        self._client.open_api_do(
            "PATCH", f"drafts/{draft_number}", self._dataset_id, json=patch_data
        )

    def close_draft(self, number: int) -> None:
        """Close the draft.

        Arguments:
            number: The draft number.

        Raises:
            StatusError: When closing the current draft.

        """
        if number == self._status.draft_number:
            raise StatusError("Closing the current draft is not allowed")

        self._close_draft(number)

    def squash_and_merge(
        self,
        title: str,
        description: str = "",
        *,
        source_branch_name: str,
        target_branch_name: Optional[str] = None,
        strategy: Optional[str] = "abort",
    ) -> int:
        """Squash and merge.

        Squash commits in source branch, then merge into target branch by creating a new draft.
        If the target branch name is not given, the draft will be based on the branch name stored
        in the dataset client. And during merging, the conflicts between branches can be resolved
        in three different strategies: "abort", "override" and "skip".

        Arguments:
            title: The draft title.
            description: The draft description.
            source_branch_name: The name of the branch to be squashed.
            target_branch_name: The target branch name of the merge operation.
            strategy: The strategy of handling the branch conflict. There are three options:

                1. "abort": abort the opetation;
                2. "override": the squashed branch will override the target branch;
                3. "skip": keep the origin branch.

        Raises:
            StatusError: When squashing and merging without basing on a branch.

        Returns:
            The draft number of the new draft.

        """
        if not target_branch_name:
            target_branch_name = self._status.branch_name
            if not target_branch_name:
                raise StatusError(
                    message="Squash and merge without basing on a branch is not allowed"
                )
            self._status.check_authority_for_commit()

        post_data: Dict[str, Any] = {
            "title": title,
            "sourceBranchName": source_branch_name,
            "targetBranchName": target_branch_name,
            "strategy": strategy,
        }

        if description:
            post_data["description"] = description

        response = self._client.open_api_do(
            "POST", "squashAndMerge", self._dataset_id, json=post_data
        )
        draft_number: int = response.json()["draftNumber"]

        self._status.checkout(draft_number=draft_number)
        self._status.branch_name = target_branch_name
        return draft_number

    def get_commit(self, revision: Optional[str] = None) -> Commit:
        """Get the certain commit with the given revision.

        Get the certain commit with the given revision. If the revision is not given,
        get the commit based on the commit id stored in the dataset client.

        Arguments:
            revision: The information to locate the specific commit, which can be the commit id,
                the branch name, or the tag name.
                If is not given, get the current commit.

        Returns:
            The :class:`.Commit` instance with the given revision.

        Raises:
            TypeError: When the given revision is illegal.
            ResourceNotExistError: When the required commit does not exist.

        """
        if revision is None:
            self._status.check_authority_for_commit()
            revision = self._status.commit_id

        if not revision:
            raise TypeError("The given revision is illegal")

        try:
            commit = next(self._generate_commits(revision))
        except StopIteration as error:
            raise ResourceNotExistError(resource="commit", identification=revision) from error

        return commit

    def list_commits(self, revision: Optional[str] = None) -> PagingList[Commit]:
        """List the commits.

        Arguments:
            revision: The information to locate the specific commit, which can be the commit id,
                the branch name, or the tag name.
                If is given, list the commits before the given commit.
                If is not given, list the commits before the current commit.

        Raises:
            TypeError: When the given revision is illegal.

        Returns:
            The PagingList of :class:`commits<.Commit>`.

        """
        if revision is None:
            if self._status.is_draft:
                revision = self._status.branch_name
            else:
                revision = self._status.commit_id

        if not revision:
            raise TypeError("The given revision is illegal")

        return PagingList(
            lambda offset, limit: self._generate_commits(
                revision, offset, limit  # type: ignore[arg-type]
            ),
            128,
        )

    def create_branch(self, name: str, revision: Optional[str] = None) -> None:
        """Create a branch.

        Create a branch based on a commit with the given revision. If the revision is not given,
        create a branch based on the commit id stored in the dataset client.
        Then the dataset client will change the status to "commit"
        and store the branch name and the commit id.

        Arguments:
            name: The branch name.
            revision: The information to locate the specific commit, which can be the commit id,
                the branch name, or the tag name.
                If the revision is not given, create the branch based on the current commit.

        """
        if not revision:
            self._status.check_authority_for_commit()
        else:
            self.checkout(revision=revision)
        revision = self._status.commit_id

        post_data: Dict[str, Any] = {"name": name, "commit": revision}
        self._client.open_api_do("POST", "branches", self._dataset_id, json=post_data)

        self._status.branch_name = name

    def get_branch(self, name: str) -> Branch:
        """Get the branch with the given name.

        Arguments:
            name: The required branch name.

        Returns:
            The :class:`.Branch` instance with the given name.

        Raises:
            TypeError: When the given branch is illegal.
            ResourceNotExistError: When the required branch does not exist.

        """
        if not name:
            raise TypeError("The given branch name is illegal")

        try:
            branch = next(self._generate_branches(name))
        except StopIteration as error:
            raise ResourceNotExistError(resource="branch", identification=name) from error

        return branch

    def list_branches(self) -> PagingList[Branch]:
        """List the information of branches.

        Returns:
            The PagingList of :class:`branches<.Branch>`.

        """
        return PagingList(lambda offset, limit: self._generate_branches(None, offset, limit), 128)

    def delete_branch(self, name: str) -> None:
        """Delete a branch.

        Delete the branch with the given branch name. Note that deleting the branch with the name
        which is stored in the current dataset client is not allowed.

        Arguments:
            name: The name of the branch to be deleted.

        Raises:
            StatusError: When deleting the current branch.

        """
        if name == self._status.branch_name:
            raise StatusError("Deleting the current branch is not allowed")

        self._delete_branch(name)

    def create_tag(self, name: str, revision: Optional[str] = None) -> None:
        """Create a tag for a commit.

        Create a tag for a commit with the given revision. If the revision is not given,
        create a tag based on the commit id stored in the dataset client.

        Arguments:
            name: The tag name to be created for the specific commit.
            revision: The information to locate the specific commit, which can be the commit id,
                the branch name, or the tag name.
                If the revision is not given, create the tag for the current commit.

        """
        if not revision:
            self._status.check_authority_for_commit()
            revision = self._status.commit_id

        post_data: Dict[str, Any] = {"commit": revision, "name": name}

        self._client.open_api_do("POST", "tags", self._dataset_id, json=post_data)

    def get_tag(self, name: str) -> Tag:
        """Get the certain tag with the given name.

        Arguments:
            name: The required tag name.

        Returns:
            The :class:`.Tag` instance with the given name.

        Raises:
            TypeError: When the given tag is illegal.
            ResourceNotExistError: When the required tag does not exist.

        """
        if not name:
            raise TypeError("The given tag name is illegal")

        try:
            tag = next(self._generate_tags(name))
        except StopIteration as error:
            raise ResourceNotExistError(resource="tag", identification=name) from error

        return tag

    def list_tags(self) -> PagingList[Tag]:
        """List the information of tags.

        Returns:
            The PagingList of :class:`tags<.Tag>`.

        """
        return PagingList(lambda offset, limit: self._generate_tags(None, offset, limit), 128)

    def delete_tag(self, name: str) -> None:
        """Delete a tag.

        Arguments:
            name: The tag name to be deleted for the specific commit.

        """
        delete_data: Dict[str, Any] = {"name": name}

        self._client.open_api_do("DELETE", "tags", self._dataset_id, json=delete_data)



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of OxfordIIITPet dataset."""

import os
from typing import Any, List

from tensorbay.dataset import Data, Dataset
from tensorbay.label import Classification, LabeledBox2D, SemanticMask
from tensorbay.opendataset._utility.glob import glob

try:
    import xmltodict
except ModuleNotFoundError:
    from tensorbay.opendataset._utility.mocker import xmltodict  # pylint:disable=ungrouped-imports

DATASET_NAME = "OxfordIIITPet"


def OxfordIIITPet(path: str) -> Dataset:
    """`OxfordIIITPet <https://www.robots.ox.ac.uk/~vgg/data/pets/>`_ dataset.

    The file structure should be like::

        <path>
            annotations/
                trimaps/
                    Bombay_113.png
                    Bombay_114.png
                    ...
                xmls/
                    Birman_174.xml
                    Birman_175.xml
                    ...
                list.txt
                test.txt
                trainval.txt
                README
            images/
                Bombay_117.jpg
                Bombay_118.jpg
                ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class: `~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))
    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))
    trainval_segment = dataset.create_segment("trainval")
    test_segment = dataset.create_segment("test")
    annotation_path = os.path.join(root_path, "annotations")
    for image_path in glob(os.path.join(root_path, "images", "*.jpg")):
        stem = os.path.splitext(os.path.basename(image_path))[0]
        name = "Cat" if stem.istitle() else "Dog"
        category, num = stem.rsplit("_", 1)
        data = Data(image_path, target_remote_path=f"{category}_{num.zfill(3)}.jpg")
        label = data.label
        label.classification = Classification(category=f"{name}.{category}")
        label.semantic_mask = SemanticMask(os.path.join(annotation_path, "trimaps", f"{stem}.png"))
        xml_path = os.path.join(annotation_path, "xmls", f"{stem}.xml")
        if os.path.exists(xml_path):
            label.box2d = _get_box_label(xml_path)
            trainval_segment.append(data)
        else:
            test_segment.append(data)
    return dataset


def _get_box_label(file_path: str) -> List[LabeledBox2D]:
    with open(file_path, "r", encoding="utf-8") as fp:
        labels: Any = xmltodict.parse(fp.read())

    objects = labels["annotation"]["object"]
    if not isinstance(objects, list):
        objects = [objects]
    box2d = []
    for obj in objects:
        bndbox = obj["bndbox"]
        box2d.append(
            LabeledBox2D(
                float(bndbox["xmin"]),
                float(bndbox["ymin"]),
                float(bndbox["xmax"]),
                float(bndbox["ymax"]),
            )
        )
    return box2d



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Graviti python SDK."""

from tensorbay.__version__ import __version__
from tensorbay.client import GAS

__all__ = ["__version__", "GAS"]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""TypeEnum, TypeMixin and TypeRegister.

:class:`TypeEnum` is a superclass for enumeration classes that need to create a mapping with class.

:class:`TypeMixin` is a superclass for the class which needs to link with :class:`TypeEnum`.

:class:`TypeRegister` is a decorator, which is used for registering
:class:`TypeMixin` to :class:`TypeEnum`.

"""

from enum import Enum
from typing import Any, Dict, Generic, Type, TypeVar


class TypeEnum(Enum):
    """TypeEnum is a superclass for enumeration classes that need to create a mapping with class.

    The 'type' property is used for getting the corresponding class of the enumeration.

    """

    __registry__: Dict["TypeEnum", Type[Any]] = {}

    def __init_subclass__(cls) -> None:
        cls.__registry__ = {}

    @property
    def type(self) -> Type[Any]:
        """Get the corresponding class.

        Returns:
            The corresponding class.

        """
        return self.__registry__[self]


_T = TypeVar("_T", bound=TypeEnum)


class TypeMixin(Generic[_T]):
    """TypeMixin is a superclass for the class which needs to link with TypeEnum.

    It provides the class variable 'TYPE' to access the corresponding TypeEnum.

    """

    _enum: _T

    @property
    def enum(self) -> _T:
        """Get the corresponding TypeEnum.

        Returns:
            The corresponding TypeEnum.

        """
        return self._enum


_S = TypeVar("_S", bound=TypeMixin[Any])


class TypeRegister(Generic[_T]):
    """TypeRegister is a decorator, which is used for registering TypeMixin to TypeEnum.

    Arguments:
        enum: The corresponding :class:`TypeEnum` of the :class:`TypeMixin`.

    """

    def __init__(self, enum: _T) -> None:
        self._enum = enum

    def __call__(self, class_: Type[_S]) -> Type[_S]:
        """Call the TypeRegister as a function.

        Arguments:
            class_: The :class:`TypeEnum` of the :class:`TypeMixin` to be registered.

        Returns:
            The :class:`TypeEnum` of the :class:`TypeMixin`.

        """
        class_._enum = self._enum
        self._enum.__registry__[self._enum] = class_
        return class_



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""SubcatalogBase.

:class:`Subcatalogbase` is the base class for different types of subcatalogs,
which defines the basic concept of Subcatalog.

A subcatalog class extends :class:`SubcatalogBase` and needed :class:`SubcatalogMixin` classes.

"""

from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar, Union

from tensorbay.utility import AttrsMixin, ReprMixin, ReprType, attr, common_loads


class SubcatalogBase(ReprMixin, AttrsMixin):
    """This is the base class for different types of subcatalogs.

    It defines the basic concept of Subcatalog, which is the collection of the labels information.
    Subcatalog contains the features, fields and specific definitions of the labels.

    The Subcatalog format varies by label type.

    Arguments:
        description: The description of the entire subcatalog.

    Attributes:
        description: The description of the entire subcatalog.

    """

    _T = TypeVar("_T", bound="SubcatalogBase")

    _repr_type = ReprType.INSTANCE
    _repr_attrs: Tuple[str, ...] = (
        "is_sample",
        "sample_rate",
        "is_tracking",
        "keypoints",
        "category_delimiter",
        "categories",
        "attributes",
        "lexicon",
    )
    description: str = attr(default="")

    def __init__(self, description: str = "") -> None:
        self.description = description

    @classmethod
    def loads(cls: Type[_T], contents: Dict[str, Any]) -> _T:
        """Loads a subcatalog from a dict containing the information of the subcatalog.

        Arguments:
            contents: A dict containing the information of the subcatalog.

        Returns:
            The loaded :class:`SubcatalogBase` object.

        """
        return common_loads(cls, contents)

    def dumps(self) -> Dict[str, Any]:
        """Dumps all the information of the subcatalog into a dict.

        Returns:
            A dict containing all the information of the subcatalog.

        """
        return self._dumps()


AttributeType = Dict[str, Union[str, int, float, bool, List[Union[str, int, float, bool]]]]


class _LabelBase(AttrsMixin, ReprMixin):
    """This class defines the basic concept of label.

    :class:`_LabelBase` is the most basic label level in the TensorBay dataset structure,
    and is the base class for the following various types of label classes.

    Each :class:`_LabelBase` object
    contains one annotaion of a :class:`~tensorbay.dataset.data.Data` object.

    Arguments:
        category: The category of the label.
        attributes: The attributes of the label.
        instance: The instance id of the label.

    Attributes:
        category: The category of the label.
        attributes: The attributes of the label.
        instance: The instance id of the label.

    """

    _repr_attrs: Tuple[str, ...] = ("category", "attributes", "instance")

    category: str = attr(is_dynamic=True)
    attributes: AttributeType = attr(is_dynamic=True)
    instance: str = attr(is_dynamic=True)

    def __init__(
        self,
        category: Optional[str] = None,
        attributes: Optional[AttributeType] = None,
        instance: Optional[str] = None,
    ):
        if category:
            self.category = category
        if attributes:
            self.attributes = attributes
        if instance:
            self.instance = instance

    def dumps(self) -> Dict[str, Any]:
        """Dumps the label into a dict.

        Returns:
            A dict containing all the information of the label.
            See dict format details in ``dumps()`` of different label classes .

        """
        return self._dumps()



#!/usr/bin/env python3
#
# Copytright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the OxfordIIITPet dataset."""

from tensorbay.opendataset.OxfordIIITPet.loader import OxfordIIITPet

__all__ = ["OxfordIIITPet"]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Implementation of gas branch."""

import click

from tensorbay.cli.tbrn import TBRN, TBRNType
from tensorbay.cli.utility import (
    ContextInfo,
    error,
    exception_handler,
    get_dataset_client,
    shorten,
    sort_branches_or_tags,
)
from tensorbay.client.gas import DatasetClientType


@exception_handler
def _implement_branch(  # pylint: disable=too-many-arguments
    obj: ContextInfo, tbrn: str, name: str, verbose: bool, is_delete: bool, sort_key: str
) -> None:
    tbrn_info = TBRN(tbrn=tbrn)
    if tbrn_info.type != TBRNType.DATASET:
        error(f'To operate a branch, "{tbrn_info}" must be a dataset')

    gas = obj.get_gas()
    dataset_client = get_dataset_client(gas, tbrn_info)

    if is_delete:
        _delete_branch(dataset_client, tbrn_info)
        return

    if name:
        _create_branch(dataset_client, name)
    else:
        _list_branches(dataset_client, verbose, sort_key)


def _create_branch(dataset_client: DatasetClientType, name: str) -> None:
    if dataset_client.status.is_draft:
        error("Branch cannot be created from a draft")

    dataset_client.create_branch(name)
    branch_tbrn = TBRN(dataset_client.name, revision=name).get_colored_tbrn()
    click.echo(f'Successfully created branch "{branch_tbrn}"')


def _list_branches(dataset_client: DatasetClientType, verbose: bool, sort_key: str) -> None:
    branches = sort_branches_or_tags(sort_key, dataset_client.list_branches())
    if not verbose:
        for branch in branches:
            click.echo(branch.name)
    else:
        name_length = max(len(branch.name) for branch in branches)
        for branch in branches:
            click.echo(f"{branch.name:{name_length}} {shorten(branch.commit_id)} {branch.title}")


def _delete_branch(dataset_client: DatasetClientType, tbrn_info: TBRN) -> None:
    if not tbrn_info.revision:
        error(f'To delete a branch, "{tbrn_info.get_tbrn()}" must have a branch name')

    dataset_client._delete_branch(tbrn_info.revision)  # pylint: disable=protected-access
    click.echo(f'Successfully deleted branch "{tbrn_info.get_colored_tbrn()}"')



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Box2D, Box3D.

:class:`Box2D` contains the information of a 2D bounding box, such as the coordinates,
width and height.
It provides :meth:`Box2D.iou` to calculate the intersection over union of two 2D boxes.

:class:`Box3D` contains the information of a 3D bounding box such as the transform,
translation, rotation and size.
It provides :meth:`Box3D.iou` to calculate the intersection over union of two 3D boxes.

"""

import math
import warnings
from typing import Dict, Iterable, Mapping, Optional, Tuple, Type, TypeVar

from tensorbay.geometry.transform import Transform3D
from tensorbay.geometry.vector import Vector2D, Vector3D
from tensorbay.utility import MatrixType, ReprMixin, ReprType, UserSequence, common_loads

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from quaternion import quaternion


_B2 = TypeVar("_B2", bound="Box2D")
_B3 = TypeVar("_B3", bound="Box3D")


class Box2D(UserSequence[float]):
    """This class defines the concept of Box2D.

    :class:`Box2D` contains the information of a 2D bounding box, such as the coordinates,
    width and height.
    It provides :meth:`Box2D.iou` to calculate the intersection over union of two 2D boxes.

    Arguments:
        xmin: The x coordinate of the top-left vertex of the 2D box.
        ymin: The y coordinate of the top-left vertex of the 2D box.
        xmax: The x coordinate of the bottom-right vertex of the 2D box.
        ymax: The y coordinate of the bottom-right vertex of the 2D box.

    Examples:
        >>> Box2D(1, 2, 3, 4)
        Box2D(1, 2, 3, 4)

    """

    _repr_type = ReprType.INSTANCE

    _LENGTH = 4

    def __init__(
        self,
        xmin: float,
        ymin: float,
        xmax: float,
        ymax: float,
    ) -> None:
        if xmin >= xmax or ymin >= ymax:
            self._data = (0.0,) * Box2D._LENGTH
        else:
            self._data = (xmin, ymin, xmax, ymax)

    def __len__(self) -> int:
        return Box2D._LENGTH

    def __and__(self, other: "Box2D") -> "Box2D":
        """Calculate the intersect box of two boxes.

        Arguments:
            other: The other box.

        Returns:
            The intersect box of the two boxes.

        """
        try:
            xmin = max(self._data[0], other._data[0])
            ymin = max(self._data[1], other._data[1])
            xmax = min(self._data[2], other._data[2])
            ymax = min(self._data[3], other._data[3])
            return Box2D(xmin, ymin, xmax, ymax)
        except (TypeError, IndexError, AttributeError):
            return NotImplemented

    def _loads(self, contents: Mapping[str, float]) -> None:
        self._data = (contents["xmin"], contents["ymin"], contents["xmax"], contents["ymax"])

    def _repr_head(self) -> str:
        """Return basic information of the Box2D.

        Returns:
            Basic information of the Box2D.

        """
        return f"{self.__class__.__name__}{self._data}"

    @staticmethod
    def iou(box1: "Box2D", box2: "Box2D") -> float:
        """Calculate the intersection over union of two 2D boxes.

        Arguments:
            box1: A 2D box.
            box2: A 2D box.

        Returns:
            The intersection over union between the two input boxes.

        Examples:
            >>> box2d_1 = Box2D(1, 2, 3, 4)
            >>> box2d_2 = Box2D(2, 2, 3, 4)
            >>> Box2D.iou(box2d_1, box2d_2)
            0.5

        """
        area1 = box1.area()
        area2 = box2.area()
        intersect_box = box1 & box2
        intersect = intersect_box.area()
        union = area1 + area2 - intersect
        return intersect / union

    @classmethod
    def from_xywh(cls: Type[_B2], x: float, y: float, width: float, height: float) -> _B2:
        """Create a :class:`Box2D` instance from the top-left vertex and the width and the height.

        Arguments:
            x: X coordinate of the top left vertex of the box.
            y: Y coordinate of the top left vertex of the box.
            width: Length of the box along the x axis.
            height: Length of the box along the y axis.

        Returns:
            The created :class:`Box2D` instance.

        Examples:
            >>> Box2D.from_xywh(1, 2, 3, 4)
            Box2D(1, 2, 4, 6)

        """
        return cls(x, y, x + width, y + height)

    @classmethod
    def loads(cls: Type[_B2], contents: Mapping[str, float]) -> _B2:
        """Load a :class:`Box2D` from a dict containing coordinates of the 2D box.

        Arguments:
            contents: A dict containing coordinates of a 2D box.

        Returns:
            The loaded :class:`Box2D` object.

        Examples:
            >>> contents = {"xmin": 1.0, "ymin": 2.0, "xmax": 3.0, "ymax": 4.0}
            >>> Box2D.loads(contents)
            Box2D(1.0, 2.0, 3.0, 4.0)

        """
        return common_loads(cls, contents)

    @property
    def xmin(self) -> float:
        """Return the minimum x coordinate.

        Returns:
            Minimum x coordinate.

        Examples:
            >>> box2d = Box2D(1, 2, 3, 4)
            >>> box2d.xmin
            1

        """
        return self._data[0]

    @property
    def ymin(self) -> float:
        """Return the minimum y coordinate.

        Returns:
            Minimum y coordinate.

        Examples:
            >>> box2d = Box2D(1, 2, 3, 4)
            >>> box2d.ymin
            2

        """
        return self._data[1]

    @property
    def xmax(self) -> float:
        """Return the maximum x coordinate.

        Returns:
            Maximum x coordinate.

        Examples:
            >>> box2d = Box2D(1, 2, 3, 4)
            >>> box2d.xmax
            3

        """
        return self._data[2]

    @property
    def ymax(self) -> float:
        """Return the maximum y coordinate.

        Returns:
            Maximum y coordinate.

        Examples:
            >>> box2d = Box2D(1, 2, 3, 4)
            >>> box2d.ymax
            4

        """
        return self._data[3]

    @property
    def tl(self) -> Vector2D:  # pylint: disable=invalid-name
        """Return the top left point.

        Returns:
            The top left point.

        Examples:
            >>> box2d = Box2D(1, 2, 3, 4)
            >>> box2d.tl
            Vector2D(1, 2)

        """
        return Vector2D(self._data[0], self._data[1])

    @property
    def br(self) -> Vector2D:  # pylint: disable=invalid-name
        """Return the bottom right point.

        Returns:
            The bottom right point.

        Examples:
            >>> box2d = Box2D(1, 2, 3, 4)
            >>> box2d.br
            Vector2D(3, 4)

        """
        return Vector2D(self._data[2], self._data[3])

    @property
    def width(self) -> float:
        """Return the width of the 2D box.

        Returns:
            The width of the 2D box.

        Examples:
            >>> box2d = Box2D(1, 2, 3, 6)
            >>> box2d.width
            2

        """
        return self._data[2] - self._data[0]

    @property
    def height(self) -> float:
        """Return the height of the 2D box.

        Returns:
            The height of the 2D box.

        Examples:
            >>> box2d = Box2D(1, 2, 3, 6)
            >>> box2d.height
            4

        """
        return self._data[3] - self._data[1]

    def dumps(self) -> Dict[str, float]:
        """Dumps a 2D box into a dict.

        Returns:
            A dict containing vertex coordinates of the box.

        Examples:
            >>> box2d = Box2D(1, 2, 3, 4)
            >>> box2d.dumps()
            {'xmin': 1, 'ymin': 2, 'xmax': 3, 'ymax': 4}

        """
        return {
            "xmin": self._data[0],
            "ymin": self._data[1],
            "xmax": self._data[2],
            "ymax": self._data[3],
        }

    def area(self) -> float:
        """Return the area of the 2D box.

        Returns:
            The area of the 2D box.

        Examples:
            >>> box2d = Box2D(1, 2, 3, 4)
            >>> box2d.area()
            4

        """
        return self.width * self.height


class Box3D(ReprMixin):
    """This class defines the concept of Box3D.

    :class:`Box3D` contains the information of a 3D bounding box such as the transform,
    translation, rotation and size.
    It provides :meth:`Box3D.iou` to calculate the intersection over union of two 3D boxes.

    Arguments:
        translation: Translation in a sequence of [x, y, z].
        rotation: Rotation in a sequence of [w, x, y, z] or numpy quaternion.
        size: Size in a sequence of [x, y, z].
        transform_matrix: A 4x4 or 3x4 transform matrix.

    Examples:
        *Initialization Method 1:* Init from size, translation and rotation.

        >>> Box3D([1, 2, 3], [0, 1, 0, 0], [1, 2, 3])
        Box3D(
          (size): Vector3D(1, 2, 3)
          (translation): Vector3D(1, 2, 3),
          (rotation): quaternion(0, 1, 0, 0),
        )

        *Initialization Method 2:* Init from size and transform matrix.

        >>> from tensorbay.geometry import Transform3D
        >>> matrix = [[1, 0, 0, 1], [0, 1, 0, 2], [0, 0, 1, 3]]
        >>> Box3D(size=[1, 2, 3], transform_matrix=matrix)
        Box3D(
          (size): Vector3D(1, 2, 3)
          (translation): Vector3D(1, 2, 3),
          (rotation): quaternion(1, -0, -0, -0),
        )

    """

    _repr_type = ReprType.INSTANCE
    _repr_attrs: Tuple[str, ...] = ("size", "translation", "rotation")

    def __init__(
        self,
        size: Iterable[float],
        translation: Iterable[float] = (0, 0, 0),
        rotation: Transform3D.RotationType = (1, 0, 0, 0),
        *,
        transform_matrix: Optional[MatrixType] = None,
    ) -> None:
        self._transform = Transform3D(translation, rotation, matrix=transform_matrix)
        self._size = Vector3D(*size)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return False

        return self._size.__eq__(other._size) and self._transform.__eq__(other._transform)

    def __rmul__(self: _B3, other: Transform3D) -> _B3:
        if isinstance(other, (Transform3D, quaternion)):
            box: _B3 = object.__new__(self.__class__)
            box._transform = other * self._transform
            box._size = self._size
            return box

        return NotImplemented  # type: ignore[unreachable]

    @staticmethod
    def _line_intersect(length1: float, length2: float, midpoint_distance: float) -> float:
        """Calculate the intersect length between two parallel lines.

        Arguments:
            length1: The length of line1.
            length2: the length of line2.
            midpoint_distance: The distance between midpoints of the two lines.

        Returns:
            The intersect length between line1 and line2.

        Examples:
            >>> Box3D._line_intersect(4, 4, 1)
            3.0

        """
        line1_min = -length1 / 2
        line1_max = length1 / 2
        line2_min = -length2 / 2 + midpoint_distance
        line2_max = length2 / 2 + midpoint_distance
        intersect_length = min(line1_max, line2_max) - max(line1_min, line2_min)
        return intersect_length if intersect_length > 0 else 0

    def _loads(self, contents: Mapping[str, Mapping[str, float]]) -> None:
        self._size = Vector3D.loads(contents["size"])
        self._transform = Transform3D.loads(contents)

    @classmethod
    def loads(cls: Type[_B3], contents: Mapping[str, Mapping[str, float]]) -> _B3:
        """Load a :class:`Box3D` from a dict containing the coordinates of the 3D box.

        Arguments:
            contents: A dict containing the coordinates of a 3D box.

        Returns:
            The loaded :class:`Box3D` object.

        Examples:
            >>> contents = {
            ...     "size": {"x": 1.0, "y": 2.0, "z": 3.0},
            ...     "translation": {"x": 1.0, "y": 2.0, "z": 3.0},
            ...     "rotation": {"w": 0.0, "x": 1.0, "y": 0.0, "z": 0.0},
            ... }
            >>> Box3D.loads(contents)
            Box3D(
              (size): Vector3D(1.0, 2.0, 3.0)
              (translation): Vector3D(1.0, 2.0, 3.0),
              (rotation): quaternion(0, 1, 0, 0),
            )

        """
        return common_loads(cls, contents)

    @classmethod
    def iou(cls, box1: "Box3D", box2: "Box3D", angle_threshold: float = 5) -> float:
        """Calculate the intersection over union between two 3D boxes.

        Arguments:
            box1: A 3D box.
            box2: A 3D box.
            angle_threshold: The threshold of the relative angles
                between two input 3d boxes in degree.

        Returns:
            The intersection over union of the two 3D boxes.

        Examples:
            >>> box3d_1 = Box3D(size=[1, 1, 1])
            >>> box3d_2 = Box3D(size=[2, 2, 2])
            >>> Box3D.iou(box3d_1, box3d_2)
            0.125

        """
        box2 = box1.transform.inverse() * box2
        if abs(math.degrees(box2.rotation.angle())) > angle_threshold:
            return 0

        intersect_size = [
            cls._line_intersect(*args) for args in zip(box1.size, box2.size, box2.translation)
        ]
        intersect = intersect_size[0] * intersect_size[1] * intersect_size[2]
        union = box1.volume() + box2.volume() - intersect
        return intersect / union

    @property
    def translation(self) -> Vector3D:
        """Return the translation of the 3D box.

        Returns:
            The translation of the 3D box.

        Examples:
            >>> box3d = Box3D(size=(1, 1, 1), translation=(1, 2, 3))
            >>> box3d.translation
            Vector3D(1, 2, 3)

        """
        return self._transform.translation

    @property
    def rotation(self) -> quaternion:
        """Return the rotation of the 3D box.

        Returns:
            The rotation of the 3D box.

        Examples:
            >>> box3d = Box3D(size=(1, 1, 1), rotation=(0, 1, 0, 0))
            >>> box3d.rotation
            quaternion(0, 1, 0, 0)

        """
        return self._transform.rotation

    @property
    def transform(self) -> Transform3D:
        """Return the transform of the 3D box.

        Returns:
            The transform of the 3D box.

        Examples:
            >>> box3d = Box3D(size=(1, 1, 1), translation=(1, 2, 3), rotation=(1, 0, 0, 0))
            >>> box3d.transform
            Transform3D(
              (translation): Vector3D(1, 2, 3),
              (rotation): quaternion(1, 0, 0, 0)
            )

        """
        return self._transform

    @property
    def size(self) -> Vector3D:
        """Return the size of the 3D box.

        Returns:
            The size of the 3D box.

        Examples:
            >>> box3d = Box3D(size=(1, 1, 1))
            >>> box3d.size
            Vector3D(1, 1, 1)

        """
        return self._size

    def volume(self) -> float:
        """Return the volume of the 3D box.

        Returns:
            The volume of the 3D box.

        Examples:
            >>> box3d = Box3D(size=(1, 2, 3))
            >>> box3d.volume()
            6

        """
        return self.size.x * self.size.y * self.size.z

    def dumps(self) -> Dict[str, Dict[str, float]]:
        """Dumps the 3D box into a dict.

        Returns:
            A dict containing translation, rotation and size information.

        Examples:
            >>> box3d = Box3D(size=(1, 2, 3), translation=(1, 2, 3), rotation=(0, 1, 0, 0))
            >>> box3d.dumps()
            {
                "translation": {"x": 1, "y": 2, "z": 3},
                "rotation": {"w": 0.0, "x": 1.0, "y": 0.0, "z": 0.0},
                "size": {"x": 1, "y": 2, "z": 3},
            }

        """
        contents = self._transform.dumps()
        contents["size"] = self.size.dumps()
        return contents



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of COVID_CT dataset."""

import os

from tensorbay.dataset import Data, Dataset
from tensorbay.label import Classification

DATASET_NAME = "COVID_CT"
_SEGMENT_TO_PATH = {
    "test_COVID": ("testCT_COVID.txt", "CT_COVID", "COVID"),
    "train_COVID": ("trainCT_COVID.txt", "CT_COVID", "COVID"),
    "val_COVID": ("valCT_COVID.txt", "CT_COVID", "COVID"),
    "test_NonCOVID": ("testCT_NonCOVID.txt", "CT_NonCOVID", "NonCOVID"),
    "train_NonCOVID": ("trainCT_NonCOVID.txt", "CT_NonCOVID", "NonCOVID"),
    "val_NonCOVID": ("valCT_NonCOVID.txt", "CT_NonCOVID", "NonCOVID"),
}


def COVID_CT(path: str) -> Dataset:
    """`COVID-CT <https://github.com/UCSD-AI4H/COVID-CT>`_ dataset.

    The file structure should be like::

        <path>
            Data-split/
                COVID/
                    testCT_COVID.txt
                    trainCT_COVID.txt
                    valCT_COVID.txt
                NonCOVID/
                    testCT_NonCOVID.txt
                    trainCT_NonCOVID.txt
                    valCT_NonCOVID.txt
            Images-processed/
                CT_COVID/
                    ...
                    2020.01.24.919183-p27-132.png
                    2020.01.24.919183-p27-133.png
                    ...
                    PIIS0140673620303603%8.png
                    ...
                CT_NonCOVID/
                    0.jpg
                    1%0.jog
                    ...
                    91%1.jpg
                    102.png
                    ...
                    2341.png

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.
    """
    root_path = os.path.abspath(os.path.expanduser(path))
    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))
    data_split_path = os.path.join(root_path, "Data-split")
    images_processed_path = os.path.join(root_path, "Images-processed")

    for segment_name, (split_filename, image_dir, category) in _SEGMENT_TO_PATH.items():
        segment = dataset.create_segment(segment_name)
        image_dir = os.path.join(images_processed_path, image_dir)
        with open(
            os.path.join(data_split_path, category, split_filename), "r", encoding="utf-8"
        ) as fp:
            for line in fp:
                image_path = os.path.join(image_dir, line.strip("\n"))
                data = Data(image_path)
                data.label.classification = Classification(category)
                segment.append(data)

    return dataset



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Implementation of gas dataset."""

import click

from tensorbay.cli.tbrn import TBRN, TBRNType
from tensorbay.cli.utility import ContextInfo, error, exception_handler


@exception_handler
def _implement_dataset(obj: ContextInfo, tbrn: str, is_delete: bool, yes: bool) -> None:
    gas = obj.get_gas()
    if not tbrn:
        if is_delete:
            error("Missing argument TBRN")
        for dataset_name in gas.list_dataset_names():
            click.echo(TBRN(dataset_name).get_tbrn())
        return

    tbrn_info = TBRN(tbrn=tbrn)
    if tbrn_info.type != TBRNType.DATASET:
        error(f'"{tbrn}" is not a dataset')
    colored_tbrn = tbrn_info.get_colored_tbrn()

    if is_delete:
        if not yes:
            click.confirm(
                f'Dataset "{colored_tbrn}" will be completely deleted.\nDo you want to continue?',
                abort=True,
            )

        gas.delete_dataset(tbrn_info.dataset_name)
        click.echo(f'Successfully deleted dataset "{colored_tbrn}"')
        return

    gas.create_dataset(tbrn_info.dataset_name)
    click.echo(f'Successfully created dataset "{colored_tbrn}"')



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of JHU-CROWD dataset."""

import os
from typing import Dict, List

from tensorbay.dataset import Data, Dataset
from tensorbay.label import Classification, LabeledBox2D
from tensorbay.opendataset._utility import glob

SEGMENT_LIST = ["train", "val", "test"]
DATASET_NAME = "JHU-CROWD"
_OCCLUSION_MAP = {1: "visible", 2: "partial-occlusion", 3: "full-occlusion"}
_WEATHER_CONDITION_MAP = {0: "no weather degradationi", 1: "fog/haze", 2: "rain", 3: "snow"}


def JHU_CROWD(path: str) -> Dataset:
    """`JHU-CROWD++ <http://www.crowd-counting.com/>`_ dataset.

    The file structure should be like::

        <path>
            train/
                images/
                    0000.jpg
                    ...
                gt/
                    0000.txt
                    ...
                image_labels.txt
            test/
            val/

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))
    for segment_name in SEGMENT_LIST:
        segment = dataset.create_segment(segment_name)
        segment_path = os.path.join(path, segment_name)
        image_root_path = os.path.join(segment_path, "images")
        image_paths = glob(os.path.join(image_root_path, "*.jpg"))

        image_labels = _load_image_labels(os.path.join(segment_path, "image_labels.txt"))
        for image_path in image_paths:
            data = Data(image_path)
            image_file = os.path.basename(image_path)
            label_file = image_file.replace("jpg", "txt")
            data.label.box2d = _load_box_labels(os.path.join(segment_path, "gt", label_file))
            data.label.classification = image_labels[os.path.splitext(image_file)[0]]
            segment.append(data)
    return dataset


def _load_box_labels(file_path: str) -> List[LabeledBox2D]:
    box_labels = []
    with open(file_path, encoding="utf-8") as fp:
        for line in fp:
            center_x, center_y, width, height, occlusion, blur = map(int, line.strip().split())
            attributes = {"occlusion-level": _OCCLUSION_MAP[occlusion], "blur-level": bool(blur)}
            box_labels.append(
                LabeledBox2D.from_xywh(
                    x=center_x - width / 2,
                    y=center_y - height / 2,
                    width=width,
                    height=height,
                    attributes=attributes,
                )
            )
    return box_labels


def _load_image_labels(file_path: str) -> Dict[str, Classification]:
    with open(file_path, encoding="utf-8") as fp:
        image_labels = {}
        for line in fp:
            img_index, count, scene, weather, distractor = line.strip().split(",")
            attributes = {
                "total-count": int(count),
                "scene-type": scene,
                "weather-condition": _WEATHER_CONDITION_MAP[int(weather)],
                "distractor": bool(int(distractor)),
            }
            image_labels[img_index] = Classification(attributes=attributes)
    return image_labels



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the WIDER FACE dataset."""

from tensorbay.opendataset.WIDER_FACE.loader import WIDER_FACE

__all__ = ["WIDER_FACE"]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Polyline2D.

:class:`Polyline2D` contains the coordinates of the vertexes of the polyline
and provides a series of methods to operate on polyline, such as
:meth:`Polyline2D.uniform_frechet_distance` and :meth:`Polyline2D.similarity`.

:class:`MultiPolyline2D` contains a list of polyline.
"""

from itertools import accumulate, count, islice, product
from sys import version_info
from typing import Any, Dict, Iterable, List, Mapping, Optional, Sequence, Tuple, Type, TypeVar

from tensorbay.geometry.point_list import MultiPointList2D, PointList2D
from tensorbay.geometry.vector import Vector2D
from tensorbay.utility import common_loads

if version_info >= (3, 8):
    from math import dist as _dist
else:
    from math import hypot

    def _dist(point1: Iterable[float], point2: Iterable[float]) -> float:  # type: ignore[misc]
        return hypot(*((p1 - p2) for p1, p2 in zip(point1, point2)))


class Polyline2D(PointList2D[Vector2D]):
    """This class defines the concept of Polyline2D.

    :class:`Polyline2D` contains the coordinates of the vertexes of the polyline
    and provides a series of methods to operate on polyline, such as
    :meth:`Polyline2D.uniform_frechet_distance` and :meth:`Polyline2D.similarity`.

    Examples:
        >>> Polyline2D([[1, 2], [2, 3]])
        Polyline2D [
          Vector2D(1, 2),
          Vector2D(2, 3)
        ]

    """

    _P = TypeVar("_P", bound="Polyline2D")

    _ElementType = Vector2D

    @staticmethod
    def _get_polyline_info(polyline: "Polyline2D") -> Tuple[Dict[str, Any], ...]:
        vectors = tuple(p1 - p2 for p1, p2 in zip(islice(polyline, 1, None), polyline))
        distances = tuple(accumulate(abs(v) for v in vectors))
        velocity = distances[-1]
        time = [d / velocity for d in distances]
        time.insert(0, 0)

        return tuple(
            {
                "index": index,
                "point": point,
                "vector": vector,
                "time": current_time,
                "last_time": last_time,
            }
            for index, vector, point, current_time, last_time in zip(
                count(), vectors, polyline, islice(time, 1, None), time
            )
        )

    @staticmethod
    def _get_insert_arg(time: float, info: Dict[str, Any]) -> Tuple[int, Vector2D]:
        ratio = (time - info["last_time"]) / (info["time"] - info["last_time"])
        insert_point = info["point"] + info["vector"] * ratio
        return (info["index"] + 1, insert_point)

    @staticmethod
    def _get_insert_args(
        polyline_info1: Iterable[Dict[str, Any]],
        polyline_info2: Iterable[Dict[str, Any]],
    ) -> Tuple[List[Tuple[int, Vector2D]], List[Tuple[int, Vector2D]]]:
        insert_points1: List[Tuple[int, Vector2D]] = []
        insert_points2: List[Tuple[int, Vector2D]] = []

        iter1 = iter(polyline_info1)
        iter2 = iter(polyline_info2)
        info1 = next(iter1)
        info2 = next(iter2)

        try:
            while True:
                time1 = info1["time"]
                time2 = info2["time"]
                if time1 < time2:
                    insert_points2.append(Polyline2D._get_insert_arg(time1, info2))
                    info1 = next(iter1)
                elif time1 > time2:
                    insert_points1.append(Polyline2D._get_insert_arg(time2, info1))
                    info2 = next(iter2)
                else:
                    info1 = next(iter1)
                    info2 = next(iter2)
        except StopIteration:
            pass

        return insert_points1, insert_points2

    @staticmethod
    def uniform_frechet_distance(
        polyline1: Sequence[Sequence[float]],
        polyline2: Sequence[Sequence[float]],
    ) -> float:
        """Compute the maximum distance between two curves if walk on a constant speed on a curve.

        Arguments:
            polyline1: The first polyline consists of multiple points.
            polyline2: The second polyline consists of multiple points.

        Returns:
            The computed distance between the two polylines.

        Examples:
            >>> polyline_1 = [[1, 1], [1, 2], [2, 2]]
            >>> polyline_2 = [[4, 5], [2, 1], [3, 3]]
            >>> Polyline2D.uniform_frechet_distance(polyline_1, polyline_2)
            3.605551275463989

        """
        # forward:
        line1 = Polyline2D(polyline1)
        line2 = Polyline2D(polyline2)

        polyline_info1 = Polyline2D._get_polyline_info(line1)
        polyline_info2 = Polyline2D._get_polyline_info(line2)

        insert_args1, insert_args2 = Polyline2D._get_insert_args(polyline_info1, polyline_info2)

        for arg in reversed(insert_args1):
            line1.insert(*arg)
        for arg in reversed(insert_args2):
            line2.insert(*arg)
        distance_forward = max(_dist(*args) for args in zip(line1, line2))

        # backward:
        line1 = Polyline2D(polyline1)
        line2_reverse = Polyline2D(reversed(polyline2))

        polyline_info2_reverse = Polyline2D._get_polyline_info(line2_reverse)

        insert_args1, insert_args2 = Polyline2D._get_insert_args(
            polyline_info1, polyline_info2_reverse
        )

        for arg in reversed(insert_args1):
            line1.insert(*arg)
        for arg in reversed(insert_args2):
            line2_reverse.insert(*arg)
        distance_reverse = max(_dist(*args) for args in zip(line1, line2_reverse))

        return min(distance_forward, distance_reverse)

    @staticmethod
    def similarity(
        polyline1: Sequence[Sequence[float]],
        polyline2: Sequence[Sequence[float]],
    ) -> float:
        """Calculate the similarity between two polylines, range from 0 to 1.

        Arguments:
            polyline1: The first polyline consists of multiple points.
            polyline2: The second polyline consisting of multiple points.

        Returns:
            The similarity between the two polylines.
            The larger the value, the higher the similarity.

        Examples:
            >>> polyline_1 = [[1, 1], [1, 2], [2, 2]]
            >>> polyline_2 = [[4, 5], [2, 1], [3, 3]]
            >>> Polyline2D.similarity(polyline_1, polyline_2)
            0.2788897449072022

        """
        min_distance = Polyline2D.uniform_frechet_distance(polyline1, polyline2)
        max_distance = max(_dist(*args) for args in product(polyline1, polyline2))
        return 1 - min_distance / max_distance

    @classmethod
    def loads(cls: Type[_P], contents: Sequence[Mapping[str, float]]) -> _P:
        """Load a :class:`Polyline2D` from a list of dict.

        Arguments:
            contents: A list of dict containing
                the coordinates of the vertexes of the polyline.

        Returns:
            The loaded :class:`Polyline2D` object.

        Examples:
            >>> polyline = Polyline2D([[1, 1], [1, 2], [2, 2]])
            >>> polyline.dumps()
            [{'x': 1, 'y': 1}, {'x': 1, 'y': 2}, {'x': 2, 'y': 2}]

        """
        return common_loads(cls, contents)


class MultiPolyline2D(MultiPointList2D[Polyline2D]):
    """This class defines the concept of MultiPolyline2D.

    :class:`MultiPolyline2D` contains a list of polylines.

    Arguments:
        polylines: A list of polylines.

    Examples:
        >>> MultiPolyline2D([[[1, 2], [2, 3]], [[3, 4], [6, 8]]])
        MultiPolyline2D [
            Polyline2D [...]
            Polyline2D [...]
            ...
        ]

    """

    _P = TypeVar("_P", bound="MultiPolyline2D")

    _ElementType = Polyline2D

    def __init__(self, polylines: Optional[Iterable[Iterable[Iterable[float]]]] = None) -> None:
        super().__init__(polylines)

    @classmethod
    def loads(cls: Type[_P], contents: Sequence[Sequence[Mapping[str, float]]]) -> _P:
        """Loads a :class:`MultiPolyline2D` from the given contents.

        Arguments:
            contents: A list of dict lists containing
                the coordinates of the vertexes of the polyline list.

        Returns:
            The loaded :class:`MultiPolyline2D` object.

        Examples:
            >>> contents = [[{'x': 1, 'y': 1}, {'x': 1, 'y': 2}, {'x': 2, 'y': 2}],
                            [{'x': 2, 'y': 3}, {'x': 3, 'y': 5}]]
            >>> multipolyline = MultiPolyline2D.loads(contents)
            >>> multipolyline
            MultiPolyline2D [
                Polyline2D [...]
                Polyline2D [...]
                ...
            ]

        """
        return common_loads(cls, contents)

    def dumps(self) -> List[List[Dict[str, float]]]:
        """Dumps a :class:`MultiPolyline2D` into a polyline list.

        Returns:
            All the information of the :class:`MultiPolyline2D`.

        Examples:
            >>> multipolyline = MultiPolyline2D([[[1, 1], [1, 2], [2, 2]], [[2, 3], [3, 5]]])
            >>> multipolyline.dumps()
            [
                [{'x': 1, 'y': 1}, {'x': 1, 'y': 2}, {'x': 2, 'y': 2}],
                [{'x': 2, 'y': 3}, {'x': 3, 'y': 5}]
            ]

        """
        return self._dumps()



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""LabeledBox2D ,LabeledBox3D, Box2DSubcatalog, Box3DSubcatalog.

:class:`Box2DSubcatalog` defines the subcatalog for 2D box type of labels.

:class:`LabeledBox2D` is the 2D bounding box type of label,
which is often used for CV tasks such as object detection.

:class:`Box3DSubcatalog` defines the subcatalog for 3D box type of labels.

:class:`LabeledBox3D` is the 3D bounding box type of label,
which is often used for object detection in 3D point cloud.

"""

import warnings
from typing import Any, Dict, Iterable, Mapping, Optional, Type, TypeVar

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from quaternion import quaternion

from tensorbay.geometry import Box2D, Box3D, Transform3D
from tensorbay.label.basic import SubcatalogBase, _LabelBase
from tensorbay.label.supports import AttributesMixin, CategoriesMixin, IsTrackingMixin
from tensorbay.utility import MatrixType, ReprType, attr_base, common_loads


class Box2DSubcatalog(SubcatalogBase, IsTrackingMixin, CategoriesMixin, AttributesMixin):
    """This class defines the subcatalog for 2D box type of labels.

    Arguments:
        is_tracking: A boolean value indicates whether the corresponding
            subcatalog contains tracking information.

    Attributes:
        description: The description of the entire 2D box subcatalog.
        categories: All the possible categories in the corresponding dataset
            stored in a :class:`~tensorbay.utility.name.NameList`
            with the category names as keys
            and the :class:`~tensorbay.label.supports.CategoryInfo` as values.
        category_delimiter: The delimiter in category values indicating parent-child relationship.
        attributes: All the possible attributes in the corresponding dataset
            stored in a :class:`~tensorbay.utility.name.NameList`
            with the attribute names as keys
            and the :class:`~tensorbay.label.attribute.AttributeInfo` as values.
        is_tracking: Whether the Subcatalog contains tracking information.

    Examples:
        *Initialization Method 1:* Init from ``Box2DSubcatalog.loads()`` method.

        >>> catalog = {
        ...     "BOX2D": {
        ...         "isTracking": True,
        ...         "categoryDelimiter": ".",
        ...         "categories": [{"name": "0"}, {"name": "1"}],
        ...         "attributes": [{"name": "gender", "enum": ["male", "female"]}],
        ...     }
        ... }
        >>> Box2DSubcatalog.loads(catalog["BOX2D"])
        Box2DSubcatalog(
          (is_tracking): True,
          (category_delimiter): '.',
          (categories): NameList [...],
          (attributes): NameList [...]
        )

        *Initialization Method 2:* Init an empty Box2DSubcatalog and then add the attributes.

        >>> from tensorbay.utility import NameList
        >>> from tensorbay.label import CategoryInfo, AttributeInfo
        >>> categories = NameList()
        >>> categories.append(CategoryInfo("a"))
        >>> attributes = NameList()
        >>> attributes.append(AttributeInfo("gender", enum=["female", "male"]))
        >>> box2d_subcatalog = Box2DSubcatalog()
        >>> box2d_subcatalog.is_tracking = True
        >>> box2d_subcatalog.category_delimiter = "."
        >>> box2d_subcatalog.categories = categories
        >>> box2d_subcatalog.attributes = attributes
        >>> box2d_subcatalog
        Box2DSubcatalog(
          (is_tracking): True,
          (category_delimiter): '.',
          (categories): NameList [...],
          (attributes): NameList [...]
        )

    """

    def __init__(self, is_tracking: bool = False) -> None:
        SubcatalogBase.__init__(self)
        IsTrackingMixin.__init__(self, is_tracking)


class LabeledBox2D(_LabelBase, Box2D):
    """This class defines the concept of 2D bounding box label.

    :class:`LabeledBox2D` is the 2D bounding box type of label,
    which is often used for CV tasks such as object detection.

    Arguments:
        xmin: The x coordinate of the top-left vertex of the labeled 2D box.
        ymin: The y coordinate of the top-left vertex of the labeled 2D box.
        xmax: The x coordinate of the bottom-right vertex of the labeled 2D box.
        ymax: The y coordinate of the bottom-right vertex of the labeled 2D box.
        category: The category of the label.
        attributes: The attributs of the label.
        instance: The instance id of the label.

    Attributes:
        category: The category of the label.
        attributes: The attributes of the label.
        instance: The instance id of the label.

    Examples:
        >>> xmin, ymin, xmax, ymax = 1, 2, 4, 4
        >>> LabeledBox2D(
        ...     xmin,
        ...     ymin,
        ...     xmax,
        ...     ymax,
        ...     category="example",
        ...     attributes={"attr": "a"},
        ...     instance="12345",
        ... )
        LabeledBox2D(1, 2, 4, 4)(
          (category): 'example',
          (attributes): {...},
          (instance): '12345'
        )

    """

    _T = TypeVar("_T", bound="LabeledBox2D")

    _repr_type = ReprType.INSTANCE
    _repr_attrs = _LabelBase._repr_attrs

    _attrs_base: Box2D = attr_base(key="box2d")

    def __init__(
        self,
        xmin: float,
        ymin: float,
        xmax: float,
        ymax: float,
        *,
        category: Optional[str] = None,
        attributes: Optional[Dict[str, Any]] = None,
        instance: Optional[str] = None,
    ):
        Box2D.__init__(self, xmin, ymin, xmax, ymax)
        _LabelBase.__init__(self, category, attributes, instance)

    @classmethod
    def from_xywh(  # pylint: disable=arguments-differ
        cls: Type[_T],
        x: float,
        y: float,
        width: float,
        height: float,
        *,
        category: Optional[str] = None,
        attributes: Optional[Dict[str, Any]] = None,
        instance: Optional[str] = None,
    ) -> _T:
        """Create a :class:`LabeledBox2D` instance from the top-left vertex, the width and height.

        Arguments:
            x: X coordinate of the top left vertex of the box.
            y: Y coordinate of the top left vertex of the box.
            width: Length of the box along the x axis.
            height: Length of the box along the y axis.
            category: The category of the label.
            attributes: The attributs of the label.
            instance: The instance id of the label.

        Returns:
            The created :class:`LabeledBox2D` instance.

        Examples:
            >>> x, y, width, height = 1, 2, 3, 4
            >>> LabeledBox2D.from_xywh(
            ...     x,
            ...     y,
            ...     width,
            ...     height,
            ...     category="example",
            ...     attributes={"key": "value"},
            ...     instance="12345",
            ... )
            LabeledBox2D(1, 2, 4, 6)(
              (category): 'example',
              (attributes): {...},
              (instance): '12345'
            )

        """
        return cls(
            x, y, x + width, y + height, category=category, attributes=attributes, instance=instance
        )

    @classmethod
    def loads(cls: Type[_T], contents: Mapping[str, Any]) -> _T:
        """Loads a LabeledBox2D from a dict containing the information of the label.

        Arguments:
            contents: A dict containing the information of the 2D bounding box label.

        Returns:
            The loaded :class:`LabeledBox2D` object.

        Examples:
            >>> contents = {
            ...     "box2d": {"xmin": 1, "ymin": 2, "xmax": 5, "ymax": 8},
            ...     "category": "example",
            ...     "attributes": {"key": "value"},
            ...     "instance": "12345",
            ... }
            >>> LabeledBox2D.loads(contents)
            LabeledBox2D(1, 2, 5, 8)(
              (category): 'example',
              (attributes): {...},
              (instance): '12345'
            )

        """
        return common_loads(cls, contents)

    def dumps(self) -> Dict[str, Any]:
        """Dumps the current 2D bounding box label into a dict.

        Returns:
            A dict containing all the information of the 2D box label.

        Examples:
            >>> xmin, ymin, xmax, ymax = 1, 2, 4, 4
            >>> labelbox2d = LabeledBox2D(
            ...     xmin,
            ...     ymin,
            ...     xmax,
            ...     ymax,
            ...     category="example",
            ...     attributes={"attr": "a"},
            ...     instance="12345",
            ... )
            >>> labelbox2d.dumps()
            {
                'category': 'example',
                'attributes': {'attr': 'a'},
                'instance': '12345',
                'box2d': {'xmin': 1, 'ymin': 2, 'xmax': 4, 'ymax': 4},
            }

        """
        return self._dumps()


class Box3DSubcatalog(SubcatalogBase, IsTrackingMixin, CategoriesMixin, AttributesMixin):
    """This class defines the subcatalog for 3D box type of labels.

    Arguments:
        is_tracking: A boolean value indicates whether the corresponding
            subcatalog contains tracking information.

    Attributes:
        description: The description of the entire 3D box subcatalog.
        categories: All the possible categories in the corresponding dataset
            stored in a :class:`~tensorbay.utility.name.NameList`
            with the category names as keys
            and the :class:`~tensorbay.label.supports.CategoryInfo` as values.
        category_delimiter: The delimiter in category values indicating parent-child relationship.
        attributes: All the possible attributes in the corresponding dataset
            stored in a :class:`~tensorbay.utility.name.NameList`
            with the attribute names as keys
            and the :class:`~tensorbay.label.attribute.AttributeInfo` as values.
        is_tracking: Whether the Subcatalog contains tracking information.

    Examples:
        *Initialization Method 1:* Init from ``Box3DSubcatalog.loads()`` method.

        >>> catalog = {
        ...     "BOX3D": {
        ...         "isTracking": True,
        ...         "categoryDelimiter": ".",
        ...         "categories": [{"name": "0"}, {"name": "1"}],
        ...         "attributes": [{"name": "gender", "enum": ["male", "female"]}],
        ...     }
        ... }
        >>> Box3DSubcatalog.loads(catalog["BOX3D"])
        Box3DSubcatalog(
          (is_tracking): True,
          (category_delimiter): '.',
          (categories): NameList [...],
          (attributes): NameList [...]
        )

        *Initialization Method 2:* Init an empty Box3DSubcatalog and then add the attributes.

        >>> from tensorbay.utility import NameList
        >>> from tensorbay.label import CategoryInfo, AttributeInfo
        >>> categories = NameList()
        >>> categories.append(CategoryInfo("a"))
        >>> attributes = NameList()
        >>> attributes.append(AttributeInfo("gender", enum=["female", "male"]))
        >>> box3d_subcatalog = Box3DSubcatalog()
        >>> box3d_subcatalog.is_tracking = True
        >>> box3d_subcatalog.category_delimiter = "."
        >>> box3d_subcatalog.categories = categories
        >>> box3d_subcatalog.attributes = attributes
        >>> box3d_subcatalog
        Box3DSubcatalog(
          (is_tracking): True,
          (category_delimiter): '.',
          (categories): NameList [...],
          (attributes): NameList [...]
        )

    """

    def __init__(self, is_tracking: bool = False) -> None:
        SubcatalogBase.__init__(self)
        IsTrackingMixin.__init__(self, is_tracking)


class LabeledBox3D(_LabelBase, Box3D):
    """This class defines the concept of 3D bounding box label.

    :class:`LabeledBox3D` is the 3D bounding box type of label,
    which is often used for object detection in 3D point cloud.

    Arguments:
        size: Size of the 3D bounding box label in a sequence of [x, y, z].
        translation: Translation of the 3D bounding box label in a sequence of [x, y, z].
        rotation: Rotation of the 3D bounding box label in a sequence of [w, x, y, z]
            or a numpy quaternion object.
        transform_matrix: A 4x4 or 3x4 transformation matrix.
        category: Category of the 3D bounding box label.
        attributes: Attributs of the 3D bounding box label.
        instance: The instance id of the 3D bounding box label.

    Attributes:
        category: The category of the label.
        attributes: The attributes of the label.
        instance: The instance id of the label.
        size: The size of the 3D bounding box.
        transform: The transform of the 3D bounding box.

    Examples:
        >>> LabeledBox3D(
        ...     size=[1, 2, 3],
        ...     translation=(1, 2, 3),
        ...     rotation=(0, 1, 0, 0),
        ...     category="example",
        ...     attributes={"key": "value"},
        ...     instance="12345",
        ... )
        LabeledBox3D(
          (size): Vector3D(1, 2, 3),
          (translation): Vector3D(1, 2, 3),
          (rotation): quaternion(0, 1, 0, 0),
          (category): 'example',
          (attributes): {...},
          (instance): '12345'
        )

    """

    _T = TypeVar("_T", bound="LabeledBox3D")

    _repr_attrs = Box3D._repr_attrs + _LabelBase._repr_attrs

    _attrs_base: Box3D = attr_base(key="box3d")

    def __init__(
        self,
        size: Iterable[float],
        translation: Iterable[float] = (0, 0, 0),
        rotation: Transform3D.RotationType = (1, 0, 0, 0),
        *,
        transform_matrix: Optional[MatrixType] = None,
        category: Optional[str] = None,
        attributes: Optional[Dict[str, Any]] = None,
        instance: Optional[str] = None,
    ):
        Box3D.__init__(self, size, translation, rotation, transform_matrix=transform_matrix)
        _LabelBase.__init__(self, category, attributes, instance)

    def __rmul__(self: _T, other: Transform3D) -> _T:
        if isinstance(other, (Transform3D, quaternion)):
            labeled_box_3d = Box3D.__rmul__(self, other)
            if hasattr(self, "category"):
                labeled_box_3d.category = self.category
            if hasattr(self, "attributes"):
                labeled_box_3d.attributes = self.attributes
            if hasattr(self, "instance"):
                labeled_box_3d.instance = self.instance
            return labeled_box_3d

        return NotImplemented  # type: ignore[unreachable]

    @classmethod
    def loads(cls: Type[_T], contents: Mapping[str, Any]) -> _T:
        """Loads a LabeledBox3D from a dict containing the information of the label.

        Arguments:
            contents: A dict containing the information of the 3D bounding box label.

        Returns:
            The loaded :class:`LabeledBox3D` object.

        Examples:
            >>> contents = {
            ...     "box3d": {
            ...         "size": {"x": 1, "y": 2, "z": 3},
            ...         "translation": {"x": 1, "y": 2, "z": 3},
            ...         "rotation": {"w": 1, "x": 0, "y": 0, "z": 0},
            ...     },
            ...     "category": "test",
            ...     "attributes": {"key": "value"},
            ...     "instance": "12345",
            ... }
            >>> LabeledBox3D.loads(contents)
            LabeledBox3D(
              (size): Vector3D(1, 2, 3),
              (translation): Vector3D(1, 2, 3),
              (rotation): quaternion(1, 0, 0, 0),
              (category): 'test',
              (attributes): {...},
              (instance): '12345'
            )

        """
        return common_loads(cls, contents)

    def dumps(self) -> Dict[str, Any]:
        """Dumps the current 3D bounding box label into a dict.

        Returns:
            A dict containing all the information of the 3D bounding box label.

        Examples:
            >>> labeledbox3d = LabeledBox3D(
            ...     size=[1, 2, 3],
            ...     translation=(1, 2, 3),
            ...     rotation=(0, 1, 0, 0),
            ...     category="example",
            ...     attributes={"key": "value"},
            ...     instance="12345",
            ... )
            >>> labeledbox3d.dumps()
            {
                'category': 'example',
                'attributes': {'key': 'value'},
                'instance': '12345',
                'box3d': {
                    'translation': {'x': 1, 'y': 2, 'z': 3},
                    'rotation': {'w': 0.0, 'x': 1.0, 'y': 0.0, 'z': 0.0},
                    'size': {'x': 1, 'y': 2, 'z': 3},
                },
            }

        """
        return self._dumps()



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""OpenDataset utility code."""

from tensorbay.opendataset._utility.coco import coco
from tensorbay.opendataset._utility.glob import glob

__all__ = ["coco", "glob"]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Method check_catalog.

:meth:`check_catalog` checks the catalog of :class:`~tensorbay.dataset.dataset.Dataset`
or :class:`~tensorbay.dataset.dataset.FusionDataset`,
including subcatalog, categories and attributes.
For :class:`~tensorbay.label.attributes.AttributeInfo`,
it finds errors in fields such as 'type', 'enum', 'range' and 'parent categories'.

"""

from typing import Iterator, Optional, Tuple

from tensorbay.healthcheck.pipeline import PipelineForIterable
from tensorbay.healthcheck.report import Error
from tensorbay.label import AttributeInfo, Catalog, CategoryInfo
from tensorbay.utility import NameList


class AttributeInfoError(Error):
    """This class defines :class:`AttributeInfoError`.

    Arguments:
        name: The name of the attribute which has error.

    """

    def __init__(self, name: str) -> None:
        self._name = name


ATTRIBUTE_INFO_PIPELINE: PipelineForIterable[
    AttributeInfo, AttributeInfoError
] = PipelineForIterable()


def check_catalog(catalog: Catalog) -> Iterator[Tuple[str, AttributeInfoError]]:
    """The health check method for :class:`~tensorbay.label.catalog.Catalog`.

    Arguments:
        catalog: The :class:`~tensorbay.label.catalog.Catalog` needs to be checked.

    Yields:
        The label type and :class:`AttributeInfoError` indicating that
        :class:`~tensorbay.label.attributes.AttributeInfo` has invalid 'type', 'enum', 'range'
        or 'parent categories' field.

    """
    for key in catalog._attrs_fields:  # pylint: disable=protected-access
        subcatalog = getattr(catalog, key, None)
        if not subcatalog:
            continue
        categories = getattr(subcatalog, "categories", None)
        if hasattr(subcatalog, "attributes"):
            attribute_info_pipeline = ATTRIBUTE_INFO_PIPELINE.copy()
            attribute_info_pipeline.register(CheckParentCategories(categories))
            for error in attribute_info_pipeline(subcatalog.attributes.values()):
                yield key, error


class InvalidTypeError(AttributeInfoError):
    """The health check class for invalid.

    This error is raised to indicate that
    :class:`~tensorbay.label.attributes.AttributeInfo` has invalid 'type' field.

    """

    def __str__(self) -> str:
        return f'AttributeInfo "{self._name}": "type" field is invalid'


@ATTRIBUTE_INFO_PIPELINE.register
def check_invalid_type(attribute_info: AttributeInfo) -> Iterator[InvalidTypeError]:
    """The health check method for invalid type.

    :class:`~tensorbay.label.attributes.AttributeInfo` 'type' field.

    Arguments:
        attribute_info: The :class:`~tensorbay.label.attributes.AttributeInfo` needs to be checked.

    Yields:
        :class:`InvalidTypeError` indicating that
        :class:`~tensorbay.label.attributes.AttributeInfo` has invalid 'type' field.

    """
    if not hasattr(attribute_info, "type"):
        return

    type_ = attribute_info.type

    if type_ == "null":
        yield InvalidTypeError(attribute_info.name)
        return

    if not isinstance(type_, list):
        return

    length = len(type_)
    if length in (0, 1):
        yield InvalidTypeError(attribute_info.name)
        return

    if len(set(type_)) != length:
        yield InvalidTypeError(attribute_info.name)


class InvalidEnumError(AttributeInfoError):
    """The health check class for invalid enum.

    This error is raised to indicate that
    :class:`~tensorbay.label.attributes.AttributeInfo` has invalid 'enum' field.

    """

    def __str__(self) -> str:
        return f'AttributeInfo "{self._name}": "enum" field is invalid'


@ATTRIBUTE_INFO_PIPELINE.register
def check_invalid_enum(attribute_info: AttributeInfo) -> Iterator[InvalidEnumError]:
    """The health check method for invalid enum.

    :class:`~tensorbay.label.attributes.AttributeInfo` 'enum' field.

    Arguments:
        attribute_info: The :class:`~tensorbay.label.attributes.AttributeInfo` needs to be checked.

    Yields:
        :class:`InvalidEnumError` indicating that
        :class:`~tensorbay.label.attributes.AttributeInfo` has invalid 'enum' field.

    """
    if not hasattr(attribute_info, "enum"):
        return

    enum = attribute_info.enum
    length = len(enum)
    if length in (0, 1):
        yield InvalidEnumError(attribute_info.name)
        return

    if len(set(enum)) != length:
        yield InvalidEnumError(attribute_info.name)


class NeitherTypeNorEnumError(AttributeInfoError):
    """The health check class for either type enum.

    This error is raised to indicate
    :class:`~tensorbay.label.attributes.AttributeInfo` has neither 'enum' nor 'type'.

    """

    def __str__(self) -> str:
        return f'AttributeInfo "{self._name}": Neither "type" nor "enum" field exists'


@ATTRIBUTE_INFO_PIPELINE.register
def check_neither_type_nor_enum(attribute_info: AttributeInfo) -> Iterator[NeitherTypeNorEnumError]:
    """The health check method for :class:`~tensorbay.label.attributes.AttributeInfo`.

    which has neither 'enum' nor 'type' field.

    Arguments:
        attribute_info: The :class:`~tensorbay.label.attributes.AttributeInfo` needs to be checked.

    Yields:
        :class:`NeitherTypeNorEnumError` indicating that
        :class:`~tensorbay.label.attributes.AttributeInfo` has neither 'enum' nor 'type' field.

    """
    if not hasattr(attribute_info, "enum") and not hasattr(attribute_info, "type"):
        yield NeitherTypeNorEnumError(attribute_info.name)


class RedundantTypeError(AttributeInfoError):
    """The health check class for redundant type error.

    This error is raised to indicate that
    :class:`~tensorbay.label.attributes.AttributeInfo` has both 'enum' and 'type'.

    """

    def __str__(self) -> str:
        return f'AttributeInfo "{self._name}": "type" field is redundant when "enum" field exists'


@ATTRIBUTE_INFO_PIPELINE.register
def check_redundant_type(attribute_info: AttributeInfo) -> Iterator[RedundantTypeError]:
    """The health check method for redundant type.

    :class:`~tensorbay.label.attributes.AttributeInfo`
    which has both 'enum' and 'type' field.

    Arguments:
        attribute_info: The :class:`~tensorbay.label.attributes.AttributeInfo` needs to be checked.

    Yields:
        :class:`RedundantTypeError` indicating that
        :class:`~tensorbay.label.attributes.AttributeInfo` has both 'enum' and 'type' field.

    """
    if hasattr(attribute_info, "enum") and hasattr(attribute_info, "type"):
        yield RedundantTypeError(attribute_info.name)


class RangeNotSupportError(AttributeInfoError):
    """The health check class for range not support error.

    This error is raised to indicate :class:`~tensorbay.label.attributes.AttributeInfo`
    has range for non number type.

    """

    def __str__(self) -> str:
        return f'AttributeInfo "{self._name}": Only "number" and "integer" type supports range'


@ATTRIBUTE_INFO_PIPELINE.register
def check_range_not_support(attribute_info: AttributeInfo) -> Iterator[RangeNotSupportError]:
    """The health check method for range not support.

    :class:`~tensorbay.label.attributes.AttributeInfo` which has range for non number type.

    Arguments:
        attribute_info: The :class:`~tensorbay.label.attributes.AttributeInfo` needs to be checked.

    Yields:
        :class:`RangeNotSupportError` indicating that
        :class:`~tensorbay.label.attributes.AttributeInfo` has range for non number type.

    """
    if not hasattr(attribute_info, "maximum") and not hasattr(attribute_info, "minimum"):
        return

    type_ = getattr(attribute_info, "type", None)
    if isinstance(type_, list):
        if "number" in type_:
            return

        if "integer" in type_:
            return

    elif type_ in ("number", "integer"):
        return

    yield RangeNotSupportError(attribute_info.name)


class InvalidRangeError(AttributeInfoError):
    """The health check class for invalid range error.

    This error is raised to indicate that
    :class:`~tensorbay.label.attributes.AttributeInfo` has invalid range.

    """

    def __str__(self) -> str:
        return f'AttributeInfo "{self._name}": Maximum is not larger than minimum'


@ATTRIBUTE_INFO_PIPELINE.register
def check_invalid_range(attribute_info: AttributeInfo) -> Iterator[InvalidRangeError]:
    """The health check method for invalid range.

    :class:`~tensorbay.label.attributes.AttributeInfo`
    which has invalid range.

    Arguments:
        attribute_info: The :class:`~tensorbay.label.attributes.AttributeInfo` needs to be checked.

    Yields:
        :class:`InvalidRangeError` indicating that
        :class:`~tensorbay.label.attributes.AttributeInfo` has invalid range.

    """
    if attribute_info.maximum is None or attribute_info.minimum is None:
        return

    if attribute_info.maximum > attribute_info.minimum:
        return

    yield InvalidRangeError(attribute_info.name)


class InvalidParentCategories(AttributeInfoError):
    """The health check class for invalid parent categories.

    This error is raised to indicate that :class:`~tensorbay.label.attributes.AttributeInfo`
    has invalid parent categories.This means the category in parent_categories
    cannot be found in Subcatalog.categories.

    Arguments:
        name: The name of the incorrect attribute.
        invalid_parent_category: The name of the incorrect parent_category.

    """

    def __init__(self, name: str, invalid_parent_category: str) -> None:
        super().__init__(name)
        self._invalid_parent_category = invalid_parent_category

    def __str__(self) -> str:
        return (
            f'AttributeInfo "{self._name}":'
            f'parent category "{self._invalid_parent_category}" is invalid'
        )


class CheckParentCategories:
    """The health check class for parent categories.

    This error is raised to indicate that :class:`~tensorbay.label.attributes.AttributeInfo`
    has invalid parent_categories.

    Arguments:
        categories: The dictionary of :class:`~tensorbay.label.supports.CategoryInfo`
            which indicates all valid parent categories.

    """

    def __init__(self, categories: Optional[NameList[CategoryInfo]]) -> None:
        self._keys = set(categories.keys()) if categories else set()

    def __call__(self, attribute_info: AttributeInfo) -> Iterator[InvalidParentCategories]:
        """The health check method for parent categories.

        :class:`~tensorbay.label.attributes.AttributeInfo`
        which has invalid parent categories.

        Arguments:
            attribute_info: :class:`~tensorbay.label.attributes.AttributeInfo` needs to be checked.

        Yields:
            :class:`InvalidParentCategories` indicating that
            :class:`~tensorbay.label.attributes.AttributeInfo` has invalid parent categories.

        """
        for parent_category in attribute_info.parent_categories:
            if parent_category not in self._keys:
                yield InvalidParentCategories(attribute_info.name, parent_category)



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Implementation of gas rm."""

import click

from tensorbay.cli.tbrn import TBRN, TBRNType
from tensorbay.cli.utility import ContextInfo, error, exception_handler, get_dataset_client


@exception_handler
def _implement_rm(obj: ContextInfo, tbrn: str, is_recursive: bool) -> None:
    gas = obj.get_gas()
    tbrn_info = TBRN(tbrn=tbrn)
    dataset_client = get_dataset_client(gas, tbrn_info, is_fusion=False)

    if tbrn_info.type not in (TBRNType.SEGMENT, TBRNType.NORMAL_FILE):
        error(f'"{tbrn}" is an invalid path to remove')

    if not tbrn_info.is_draft:
        error(f'To remove the data, "{tbrn}" must be in draft status, like "{tbrn}#1"')

    if tbrn_info.type == TBRNType.SEGMENT:
        if not is_recursive:
            error("Please use -r option to remove the whole segment")

        dataset_client.delete_segment(tbrn_info.segment_name)
    else:
        segment = dataset_client.get_segment(tbrn_info.segment_name)
        segment.delete_data(tbrn_info.remote_path)

    click.echo(f'Successfully deleted "{tbrn_info.get_colored_tbrn()}"')



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Implementation of gas commit."""

from typing import Tuple

import click

from tensorbay.cli.tbrn import TBRN, TBRNType
from tensorbay.cli.utility import (
    ContextInfo,
    edit_message,
    error,
    exception_handler,
    format_hint,
    get_dataset_client,
)

_COMMIT_HINT = """
# Please enter the commit message for your changes.
# The Default commit message is set as the draft title.
# Lines starting with '#' will be ignored.
# And an empty commit message aborts the commit.
"""


@exception_handler
def _implement_commit(obj: ContextInfo, tbrn: str, message: Tuple[str, ...]) -> None:
    gas = obj.get_gas()
    tbrn_info = TBRN(tbrn=tbrn)
    dataset_client = get_dataset_client(gas, tbrn_info)

    if tbrn_info.type != TBRNType.DATASET:
        error(f'To operate a commit, "{tbrn}" must be a dataset')
    if not tbrn_info.is_draft:
        error(f'To commit, "{tbrn}" must be in draft status, like "{tbrn}#1"')

    dataset_client.checkout(draft_number=tbrn_info.draft_number)
    draft = dataset_client.get_draft()
    hint_message = format_hint(draft.title, draft.description, _COMMIT_HINT)
    title, description = edit_message(message, hint_message, obj.config_parser)
    if not title:
        error("Aborting commit due to empty commit message")

    dataset_client.commit(title, description)
    commit_tbrn = TBRN(tbrn_info.dataset_name, revision=dataset_client.status.commit_id)
    click.echo(
        "Committed successfully: "
        f"{tbrn_info.get_colored_tbrn()} -> {commit_tbrn.get_colored_tbrn()}"
    )



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Implementation of gas auth."""

from configparser import NoSectionError
from textwrap import indent
from typing import Optional, Tuple
from urllib.parse import urljoin

import click

from tensorbay.cli.utility import (
    ContextInfo,
    error,
    exception_handler,
    form_profile_value,
    is_accesskey,
)
from tensorbay.exception import UnauthorizedError

INDENT = " " * 4


@exception_handler
def _implement_auth(  # pylint: disable=too-many-arguments
    obj: ContextInfo,
    arg1: str,
    arg2: str,
    get: bool,
    status: bool,
    unset: bool,
    is_all: bool,
) -> None:
    _check_args_and_options(arg1, arg2, get, status, unset, is_all)
    if get:
        _get_auth(obj, is_all)
        return

    if status:
        _status_auth(obj, is_all)
        return

    if unset:
        _unset_auth(obj, is_all)
        return

    arg1, arg2 = _reformat_args(arg1, arg2)

    _update_profile(obj, arg1, arg2)


def _get_auth(obj: ContextInfo, is_all: bool) -> None:
    config_parser = obj.config_parser
    if is_all:
        try:
            profiles = config_parser["profiles"]
        except KeyError:
            return
        for key, value in profiles.items():
            _echo_formatted_profile(key, value)
        return

    profile_name = obj.profile_name
    try:
        profile = config_parser["profiles"][profile_name]
    except KeyError:
        error(f'Profile "{profile_name}" does not exist.')
    _echo_formatted_profile(profile_name, profile)


def _unset_auth(obj: ContextInfo, is_all: bool) -> None:
    config_parser = obj.config_parser
    if is_all:
        config_parser.remove_section("profiles")
    else:
        try:
            removed = config_parser.remove_option("profiles", obj.profile_name)
        except NoSectionError:
            removed = False
        if not removed:
            error(f'Profile "{obj.profile_name}" does not exist.')
    obj.write_config(show_message=False)
    hint = "all" if is_all else f'"{obj.profile_name}"'
    click.echo(f"Successfully unset {hint} auth info")


def _interactive_auth(url: Optional[str] = None) -> str:
    click.secho(
        "Please visit and login to the TensorBay website to generate your AccessKey", bold=True
    )
    click.secho(
        "Note: TensorBay has multi-regional websites, "
        "please visit the corresponding website based on your location for better experience\n",
        fg="bright_cyan",
    )

    if url:
        developer_url = click.style(urljoin(url, "/tensorbay/developer"), underline=True)
        click.echo(f" > {developer_url}\n")
    else:
        url_cn = click.style("https://gas.graviti.cn/tensorbay/developer", underline=True)
        url_com = click.style("https://gas.graviti.com/tensorbay/developer", underline=True)
        click.echo(f" > {url_com} (Global site)")
        click.echo(f" > {url_cn} (Chinese site)\n")

    access_key = click.prompt(click.style("Paste your AccessKey here", bold=True)).strip()
    if not is_accesskey(access_key):
        error("Wrong accesskey format")
    return access_key  # type: ignore[no-any-return]


def _is_gas_url(arg: str) -> bool:
    return arg.startswith("https://gas.")


def _update_profile(obj: ContextInfo, arg1: str, arg2: str) -> None:
    access_key, url = (arg2, arg1) if arg2 else (arg1, arg2)
    profile_name, config_parser = obj.profile_name, obj.config_parser
    gas_client = obj.get_gas(access_key, url)
    try:
        user_info = gas_client.get_user()
    except UnauthorizedError:
        error(f"{access_key} is not a valid AccessKey")

    if not config_parser.has_section("profiles"):
        config_parser.add_section("profiles")
    config_parser["profiles"][profile_name] = form_profile_value(access_key, url)
    obj.write_config(show_message=False)

    messages = [
        f'Successfully set authentication info of "{click.style(user_info.name, bold=True)}"'
    ]
    if user_info.team:
        messages.append(f' in "{click.style(user_info.team.name, bold=True)}" team')
    if profile_name != "default":
        messages.append(f' into profile "{click.style(profile_name, bold=True)}"')
    click.echo("".join(messages))


def _check_args_and_options(  # pylint: disable=too-many-arguments
    arg1: str, arg2: str, get: bool, status: bool, unset: bool, is_all: bool
) -> None:
    if is_all and not (get or unset or status):
        error('Use "--all" option with "--get", "--unset" or "--status" option')

    if get + unset + status > 1:
        error('Use at most one of "--get", "--unset" and "--status"')

    if (get or unset or status) and (arg1 or arg2):
        error("Option requires 0 arguments")


def _echo_formatted_profile(name: str, value: str) -> None:
    formatted_value = indent(value, INDENT)
    click.echo(f"{name} = {formatted_value}\n")


def _status_auth(obj: ContextInfo, is_all: bool) -> None:
    if is_all:
        for profile_name, access_key, url in obj.generate_profiles():
            _echo_user_info(access_key, url, obj, profile_name)
        return

    access_key, url = obj.read_profile()
    _echo_user_info(access_key, url, obj)


def _reformat_args(arg1: str, arg2: str) -> Tuple[str, str]:
    if not arg1 and not arg2:
        arg1 = _interactive_auth()

    elif is_accesskey(arg1):
        if _is_gas_url(arg2):
            error('Please use "gas auth [url] [accessKey]" to specify the url and accessKey')
        if arg2:
            error(f'Redundant argument "{arg2}"')

    elif _is_gas_url(arg1):
        if not arg2:
            arg2 = _interactive_auth(arg1)
        elif not is_accesskey(arg2):
            error("Wrong accesskey format")
    else:
        error(f'Invalid argument "{arg1}"')
    return arg1, arg2


def _echo_user_info(
    access_key: str, url: str, obj: ContextInfo, profile_name: Optional[str] = None
) -> None:
    gas_client = obj.get_gas(access_key, url)
    if not profile_name:
        profile_name = obj.profile_name
    try:
        user_info = gas_client.get_user()
    except UnauthorizedError:
        error(f"{access_key} is not a valid AccessKey")

    click.echo(f"{profile_name}\n{INDENT}USER: {user_info.name}")
    team = user_info.team
    if team:
        click.echo(f"{INDENT}TEAM: {team.name}")
    click.echo(f"{INDENT}{access_key}")
    if url:
        click.echo(f"{INDENT}{url}\n")



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""NameMixin, SortedNameList and NameList.

:class:`NameMixin` is a mixin class for instance which has immutable name and mutable description.

:class:`SortedNameList` is a sorted sequence class which contains :class:`NameMixin`.
It is maintained in sorted order according to the 'name' of :class:`NameMixin`.

:class:`NameList` is a list of named elements, supports searching the element by its name.

"""

from bisect import bisect_left, bisect_right
from typing import Any, Dict, Iterable, List, Tuple, TypeVar, Union, overload

from tensorbay.utility.attr import AttrsMixin, attr
from tensorbay.utility.repr import ReprMixin
from tensorbay.utility.user import UserSequence


class NameMixin(AttrsMixin, ReprMixin):
    """A mixin class for instance which has immutable name and mutable description.

    Arguments:
        name: Name of the class.
        description: Description of the class.

    Attributes:
        name: Name of the class.

    """

    _P = TypeVar("_P", bound="NameMixin")

    _name: str = attr(key="name")
    description: str = attr(default="")

    def __init__(self, name: str, description: str = "") -> None:
        self._name = name
        self.description = description

    def _repr_head(self) -> str:
        return f'{self.__class__.__name__}("{self._name}")'

    @property
    def name(self) -> str:
        """Return name of the instance.

        Returns:
            Name of the instance.

        """
        return self._name


_T = TypeVar("_T", bound=NameMixin)


class NameList(UserSequence[_T]):
    """NameList is a list of named elements, supports searching the element by its name."""

    def __init__(self, values: Iterable[_T] = ()) -> None:
        self._data: List[_T] = []
        self._mapping: Dict[str, _T] = {}

        for value in values:
            self.append(value)

    @overload
    def __getitem__(self, index: Union[int, str]) -> _T:
        ...

    @overload
    def __getitem__(self, index: slice) -> List[_T]:
        ...

    def __getitem__(self, index: Union[int, str, slice]) -> Union[_T, List[_T]]:
        if isinstance(index, str):
            return self._mapping.__getitem__(index)

        return self._data.__getitem__(index)

    def __contains__(self, key: Any) -> bool:
        return self._mapping.__contains__(key)

    def keys(self) -> Tuple[str, ...]:
        """Get all element names.

        Returns:
            A tuple containing all elements names.

        """
        return tuple(item.name for item in self._data)

    def append(self, value: _T) -> None:
        """Append element to the end of the NameList.

        Arguments:
            value: Element to be appended to the NameList.

        Raises:
            KeyError: When the name of the appending object already exists in the NameList.

        """
        if value.name in self._mapping:
            raise KeyError(f'name "{value.name}" is duplicated')

        self._data.append(value)
        self._mapping[value.name] = value


class SortedNameList(UserSequence[_T]):
    """SortedNameList is a sorted sequence which contains element with name.

    It is maintained in sorted order according to the 'name' attr of the element.

    """

    def __init__(self) -> None:
        self._data: List[_T] = []
        self._names: List[str] = []

    @overload
    def __getitem__(self, index: Union[int, str]) -> _T:
        ...

    @overload
    def __getitem__(self, index: slice) -> List[_T]:
        ...

    def __getitem__(self, index: Union[int, str, slice]) -> Union[_T, List[_T]]:
        if isinstance(index, str):
            index = self._search(index)

        return self._data.__getitem__(index)

    def __delitem__(self, index: Union[int, str, slice]) -> None:
        if isinstance(index, str):
            index = self._search(index)

        self._data.__delitem__(index)
        self._names.__delitem__(index)

    def __contains__(self, key: Any) -> bool:
        try:
            self._search(key)
            return True
        except (KeyError, TypeError):
            return False

    def _search(self, key: str) -> int:
        index = bisect_left(self._names, key)
        if index == len(self._names) or self._names[index] != key:
            raise KeyError(key)

        return index

    def add(self, value: _T) -> None:
        """Store element in name sorted list.

        Arguments:
            value: The element needs to be added to the list.

        Raises:
            KeyError: If the name of the added value exists in the list.

        """
        name = value.name
        index = bisect_right(self._names, name)
        if index != 0 and self._names[index - 1] == name:
            raise KeyError(f'Name "{name}" already exists!')

        self._data.insert(index, value)
        self._names.insert(index, name)

    def keys(self) -> Tuple[str, ...]:
        """Get all element names.

        Returns:
            A tuple containing all elements names.

        """
        return tuple(self._names)



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of LIP dataset."""

import csv
import os
from itertools import islice
from typing import List

from tensorbay.dataset import Data, Dataset
from tensorbay.geometry import Keypoint2D
from tensorbay.label import LabeledKeypoints2D, SemanticMask
from tensorbay.utility import chunked

DATASET_NAME = "LIP"
_SEGMENT_NAMES = ("train", "val", "test")


def LIP(path: str) -> Dataset:
    """`LIP <https://github.com/Engineering-Course/LIP_SSL>`_ dataset.

    The file structure should be like::

        <path>
            Testing_images/
                testing_images/
                    315_462476.jpg
                    ...
                test_id.txt
            TrainVal_images/
                TrainVal_images/
                    train_images/
                        77_471474.jpg
                        ...
                    val_images/
                        36_453991.jpg
                        ...
                train_id.txt
                val_id.txt
            TrainVal_parsing_annotations/
                TrainVal_parsing_annotations/
                    train_segmentations/
                        77_471474.png
                        ...
                    val_segmentations/
                        36_453991.png
                        ...
            TrainVal_pose_annotations/
                lip_train_set.csv
                lip_val_set.csv

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded `~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))
    test_path = os.path.join(root_path, "Testing_images")
    trainval_image_path = os.path.join(root_path, "TrainVal_images", "TrainVal_images")
    trainval_parsing_path = os.path.join(
        root_path, "TrainVal_parsing_annotations", "TrainVal_parsing_annotations"
    )
    pose_path = os.path.join(root_path, "TrainVal_pose_annotations")

    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))

    for segment_name in _SEGMENT_NAMES:
        segment = dataset.create_segment(segment_name)
        if segment_name == "test":
            image_path = os.path.join(test_path, "testing_images")
            with open(os.path.join(test_path, "test_id.txt"), "r", encoding="utf-8") as fp:
                for stem in fp:
                    segment.append(Data(os.path.join(image_path, f"{stem.rstrip()}.jpg")))
        else:
            image_path = os.path.join(trainval_image_path, f"{segment_name}_images")
            parsing_path = os.path.join(trainval_parsing_path, f"{segment_name}_segmentations")
            with open(
                os.path.join(pose_path, f"lip_{segment_name}_set.csv"), "r", encoding="utf-8"
            ) as csvfile:
                for keypoints_info in csv.reader(csvfile):
                    segment.append(_get_data(keypoints_info, image_path, parsing_path))
    return dataset


def _get_data(keypoints_info: List[str], image_path: str, parsing_path: str) -> Data:
    stem = os.path.splitext(keypoints_info[0])[0]
    data = Data(os.path.join(image_path, f"{stem}.jpg"))
    label = data.label
    label.semantic_mask = SemanticMask(os.path.join(parsing_path, f"{stem}.png"))
    keypoints = LabeledKeypoints2D()
    for x, y, v in chunked(islice(keypoints_info, 1, None), 3):
        keypoints.append(
            Keypoint2D(float(x), float(y), 1 - int(v)) if x.isnumeric() else Keypoint2D(0, 0, 0)
        )
    label.keypoints2d = [keypoints]
    return data



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Class CloudClient.

The :class:`CloudClient` defines the initial client to interact between local and cloud platform.

"""

from typing import Any, Dict, Iterator, List, Type, TypeVar

from tensorbay.client.requests import Client
from tensorbay.dataset import AuthData
from tensorbay.utility import URL, AttrsMixin, ReprMixin, attr, camel, common_loads


class CloudClient:
    """:class:`CloudClient` defines the client to interact between local and cloud platform.

    Arguments:
        name: Name of the auth cloud storage config.
        client: The initial client to interact between local and TensorBay.

    """

    def __init__(self, name: str, client: Client) -> None:
        self._name = name
        self._client = client

    def _make_section(self, section: str) -> str:
        return f"cloud/{self._name}/{section}"

    def _get_url(self, file_path: str) -> str:
        params: Dict[str, Any] = {"filePath": file_path}

        response = self._client.open_api_do("GET", self._make_section("files/urls"), params=params)
        return response.json()["url"]  # type: ignore[no-any-return]

    def _list_files(self, path: str, limit: int = 128) -> Iterator[str]:
        params: Dict[str, Any] = {"prefix": path, "limit": limit}

        while True:
            response = self._client.open_api_do(
                "GET", self._make_section("files"), params=params
            ).json()
            yield from response["cloudFiles"]

            if not response["truncated"]:
                break
            params["marker"] = response["nextMarker"]

    def list_auth_data(self, path: str = "") -> List[AuthData]:
        """List all cloud files in the given directory as :class:`~tensorbay.dataset.data.AuthData`.

        Arguments:
            path: The directory path on the cloud platform.

        Returns:
            The list of :class:`~tensorbay.dataset.data.AuthData` of all the cloud files.

        """
        return [
            AuthData(
                cloud_path,
                url=URL.from_getter(
                    lambda c=cloud_path: self._get_url(c),
                    lambda c=cloud_path: self._get_url(c),  # type: ignore[misc]
                ),
            )
            for cloud_path in self._list_files(path)
        ]


class StorageConfig(AttrsMixin, ReprMixin):
    """This is a class stores the information of storage config.

    Arguments:
        name: The storage config name.
        file_path: Storage config path of the bucket.
        type_: Type of the storage provider, such as oss, s3, azure.
        is_graviti_storage: Whether the config is belong to graviti.

    """

    _T = TypeVar("_T", bound="StorageConfig")

    _repr_attrs = ("file_path", "type", "is_graviti_storage")

    name: str = attr()
    file_path: str = attr(key=camel)
    type: str = attr()
    is_graviti_storage: bool = attr(key=camel)

    def __init__(  # pylint: disable=too-many-arguments
        self,
        name: str,
        file_path: str,
        type_: str,
        is_graviti_storage: bool,
    ) -> None:
        self.name = name
        self.file_path = file_path
        self.type = type_
        self.is_graviti_storage = is_graviti_storage

    def _repr_head(self) -> str:
        return f'{self.__class__.__name__}("{self.name}")'

    @classmethod
    def loads(cls: Type[_T], contents: Dict[str, Any]) -> _T:
        """Loads a :class:`StorageConfig` instance from the given contents.

        Arguments:
            contents: The given dict containing the storage config::

                {
                    "name":                 <str>,
                    "filePath":             <str>,
                    "type":                 <str>,
                    "isGravitiStorage":     <boolean>
                }

        Returns:
            The loaded :class:`StorageConfig` instance.

        """
        return common_loads(cls, contents)

    def dumps(self) -> Dict[str, Any]:
        """Dumps the storage config into a dict.

        Returns:
            A dict containing all the information of the StorageConfig::

                {
                    "name":                 <str>,
                    "filePath":             <str>,
                    "type":                 <str>,
                    "isGravitiStorage":     <boolean>
                }

        """
        return self._dumps()



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of WIDER_FACE dataset."""

import os
from collections import OrderedDict
from itertools import islice
from typing import Dict, Iterator, List, Union

from tensorbay.dataset import Data, Dataset
from tensorbay.label import Classification, LabeledBox2D

DATASET_NAME = "WIDER_FACE"
_SEGMENT_LIST = {
    "test": "wider_face_test_filelist.txt",
    "train": "wider_face_train_bbx_gt.txt",
    "val": "wider_face_val_bbx_gt.txt",
}
_ATTRIBUTE_MAP_TYPE = Dict[str, List[Union[bool, str]]]


def WIDER_FACE(path: str) -> Dataset:
    """`WIDER FACE <http://shuoyang1213.me/WIDERFACE/>`_ dataset.

    The file structure should be like::

        <path>
            WIDER_train/
                images/
                    0--Parade/
                        0_Parade_marchingband_1_100.jpg
                        0_Parade_marchingband_1_1015.jpg
                        0_Parade_marchingband_1_1030.jpg
                        ...
                    1--Handshaking/
                    ...
                    59--people--driving--car/
                    61--Street_Battle/
            WIDER_val/
                ...
            WIDER_test/
                ...
            wider_face_split/
                wider_face_train_bbx_gt.txt
                wider_face_val_bbx_gt.txt

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))
    attribute_map = _get_attribute_map(dataset)
    for segment_name, label_file in _SEGMENT_LIST.items():
        segment = dataset.create_segment(segment_name)
        for data in _load_data(
            os.path.join(path, "wider_face_split", label_file), attribute_map, segment_name
        ):
            segment.append(data)
        segment.sort()
    return dataset


def _get_attribute_map(dataset: Dataset) -> _ATTRIBUTE_MAP_TYPE:
    attribute_map: _ATTRIBUTE_MAP_TYPE = OrderedDict()
    for info in dataset.catalog.box2d.attributes:
        if getattr(info, "type", None) == "boolean":
            attribute_map[info.name] = [False, True]
        else:
            attribute_map[info.name] = info.enum  # type: ignore[assignment]
    return attribute_map


def _load_data(path: str, attribute_map: _ATTRIBUTE_MAP_TYPE, segment_name: str) -> Iterator[Data]:
    """Loads the box2d and classification label in to data.

    The train and val label file context should be like::

        0--Parade/0_Parade_marchingband_1_849.jpg
        1
        449 330 122 149 0 0 0 0 0 0
        0--Parade/0_Parade_Parade_0_452.jpg
        0
        0 0 0 0 0 0 0 0 0 0
        0--Parade/0_Parade_marchingband_1_799.jpg
        21
        78 221 7 8 2 0 0 0 0 0
        78 238 14 17 2 0 0 0 0 0
        113 212 11 15 2 0 0 0 0 0
        134 260 15 15 2 0 0 0 0 0

    Arguments:
        path: The path of label file.
        attribute_map: A attribute value enum table.
        segment_name: Name of the segment.

    Yields:
        Data with loaded lables.

    """
    is_test = segment_name == "test"
    with open(path, encoding="utf-8") as fp:
        for image_path in fp:
            event, file_name = image_path.split("/")
            # translate directory name to category. like 0--Parade -> Parade
            category = "_".join(event.split("--")[1:])
            root_path = path.rsplit(os.sep, 2)[0]
            data = Data(
                os.path.join(root_path, f"WIDER_{segment_name}", "images", event, file_name)
            )
            data.label.classification = Classification(category)
            if not is_test:
                label_num = int(fp.readline())

                # when the label num is 0, a line of "0 0 0 0 0 0 0 0 0 0" also given
                if label_num == 0:
                    fp.readline()
                data.label.box2d = []
                for line in islice(fp, label_num):
                    labels = line.strip().split()
                    attributes = {
                        key: mapping[int(value)]
                        for (key, mapping), value in zip(attribute_map.items(), labels[4:10])
                    }
                    data.label.box2d.append(
                        LabeledBox2D.from_xywh(
                            x=int(labels[0]),
                            y=int(labels[1]),
                            width=int(labels[2]),
                            height=int(labels[3]),
                            attributes=attributes,
                        )
                    )

            yield data



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Class Status."""


from typing import Any, Dict, Optional

from tensorbay.exception import StatusError


class Status:
    """This class defines the basic concept of the status.

    Arguments:
        branch_name: The branch name.
        draft_number: The draft number (if the status is draft).
        commit_id: The commit ID (if the status is commit).

    """

    def __init__(
        self,
        branch_name: Optional[str] = None,
        *,
        draft_number: Optional[int] = None,
        commit_id: Optional[str] = None,
    ) -> None:

        self.branch_name = branch_name

        if commit_id is None and draft_number is None:
            raise StatusError(
                message="Neither commit id nor draft number is given, please give one"
            )
        if commit_id is not None and draft_number is not None:
            raise StatusError(
                message="Both commit id and draft number are given, please only give one"
            )

        self._draft_number = draft_number
        self._commit_id = commit_id

    @property
    def is_draft(self) -> bool:
        """Return whether the status is draft, True for draft, False for commit.

        Returns:
            whether the status is draft, True for draft, False for commit.

        """
        return bool(self._draft_number)

    @property
    def draft_number(self) -> Optional[int]:
        """Return the draft number.

        Returns:
            The draft number.

        """
        return self._draft_number

    @property
    def commit_id(self) -> Optional[str]:
        """Return the commit ID.

        Returns:
            The commit ID.

        """
        return self._commit_id

    def get_status_info(self) -> Dict[str, Any]:
        """Get the dict containing the draft number or commit ID.

        Returns:
            A dict containing the draft number or commit ID.

        """
        if self.is_draft:
            return {"draftNumber": self._draft_number}
        return {"commit": self._commit_id}

    def check_authority_for_commit(self) -> None:
        """Check whether the status is a legal commit.

        Raises:
            StatusError: When the status is not a legal commit.

        """
        if self._commit_id is None or self._draft_number is not None:
            raise StatusError(is_draft=self.is_draft)

    def check_authority_for_draft(self) -> None:
        """Check whether the status is a legal draft.

        Raises:
            StatusError: When the status is not a legal draft.

        """
        if self._draft_number is None or self._commit_id is not None:
            raise StatusError(is_draft=self.is_draft)

    def checkout(self, commit_id: Optional[str] = None, draft_number: Optional[int] = None) -> None:
        """Checkout to commit or draft.

        Arguments:
            commit_id: The commit ID.
            draft_number: The draft number.

        """
        self._draft_number = draft_number
        self._commit_id = commit_id



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the LISA traffic light dataset."""

from tensorbay.opendataset.LISATrafficLight.loader import LISATrafficLight

__all__ = ["LISATrafficLight"]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Classification.

:class:`ClassificationSubcatalog` defines the subcatalog for classification type of labels.

:class:`Classification` defines the concept of classification label,
which can apply to different types of data, such as images and texts.

"""

from typing import Any, Dict, Optional, Type, TypeVar

from tensorbay.label.basic import SubcatalogBase, _LabelBase
from tensorbay.label.supports import AttributesMixin, CategoriesMixin
from tensorbay.utility import common_loads


class ClassificationSubcatalog(SubcatalogBase, CategoriesMixin, AttributesMixin):
    """This class defines the subcatalog for classification type of labels.

    Attributes:
        description: The description of the entire classification subcatalog.
        categories: All the possible categories in the corresponding dataset
            stored in a :class:`~tensorbay.utility.name.NameList`
            with the category names as keys
            and the :class:`~tensorbay.label.supports.CategoryInfo` as values.
        category_delimiter: The delimiter in category values indicating parent-child relationship.
        attributes: All the possible attributes in the corresponding dataset
            stored in a :class:`~tensorbay.utility.name.NameList`
            with the attribute names as keys
            and the :class:`~tensorbay.label.attribute.AttributeInfo` as values.

    Examples:
        *Initialization Method 1:* Init from ``ClassificationSubcatalog.loads()`` method.

        >>> catalog = {
        ...     "CLASSIFICATION": {
        ...         "categoryDelimiter": ".",
        ...         "categories": [
        ...             {"name": "a"},
        ...             {"name": "b"},
        ...         ],
        ...          "attributes": [{"name": "gender", "enum": ["male", "female"]}],
        ...     }
        ... }
        >>> ClassificationSubcatalog.loads(catalog["CLASSIFICATION"])
        ClassificationSubcatalog(
          (category_delimiter): '.',
          (categories): NameList [...],
          (attributes): NameList [...]
        )

        *Initialization Method 2:* Init an empty ClassificationSubcatalog
        and then add the attributes.

        >>> from tensorbay.utility import NameList
        >>> from tensorbay.label import CategoryInfo, AttributeInfo, KeypointsInfo
        >>> categories = NameList()
        >>> categories.append(CategoryInfo("a"))
        >>> attributes = NameList()
        >>> attributes.append(AttributeInfo("gender", enum=["female", "male"]))
        >>> classification_subcatalog = ClassificationSubcatalog()
        >>> classification_subcatalog.category_delimiter = "."
        >>> classification_subcatalog.categories = categories
        >>> classification_subcatalog.attributes = attributes
        >>> classification_subcatalog
        ClassificationSubcatalog(
          (category_delimiter): '.',
          (categories): NameList [...],
          (attributes): NameList [...]
        )

    """


class Classification(_LabelBase):
    """This class defines the concept of classification label.

    :class:`Classification` is the classification type of label,
    which applies to different types of data, such as images and texts.

    Arguments:
        category: The category of the label.
        attributes: The attributes of the label.

    Attributes:
        category: The category of the label.
        attributes: The attributes of the label.

    Examples:
        >>> Classification(category="example", attributes={"attr": "a"})
        Classification(
          (category): 'example',
          (attributes): {...}
        )

    """

    _T = TypeVar("_T", bound="Classification")

    _repr_attrs = ("category", "attributes")

    def __init__(
        self,
        category: Optional[str] = None,
        attributes: Optional[Dict[str, Any]] = None,
    ):
        _LabelBase.__init__(self, category, attributes)

    @classmethod
    def loads(cls: Type[_T], contents: Dict[str, Any]) -> _T:
        """Loads a Classification label from a dict containing the label information.

        Arguments:
            contents: A dict containing the information of the classification label.

        Returns:
            The loaded :class:`Classification` object.

        Examples:
            >>> contents = {"category": "example", "attributes": {"key": "value"}}
            >>> Classification.loads(contents)
            Classification(
              (category): 'example',
              (attributes): {...}
            )

        """
        return common_loads(cls, contents)



#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#
# pylint: disable=invalid-name

"""Dataloaders of BDD100K dataset and BDD100K_10K dataset."""

import json
import os
from typing import Any, Dict, List
from warnings import warn

import numpy as np

from tensorbay.dataset import Data, Dataset
from tensorbay.label import (
    Classification,
    InstanceMask,
    LabeledBox2D,
    LabeledPolygon,
    LabeledPolyline2D,
    PanopticMask,
    SemanticMask,
)
from tensorbay.opendataset._utility import glob

try:
    from PIL import Image
except ModuleNotFoundError:
    from tensorbay.opendataset._utility.mocker import Image  # pylint:disable=ungrouped-imports

DATASET_NAMES = {
    "100k": "BDD100K",
    "10k": "BDD100K_10K",
}
_SEGMENT_NAMES = ("train", "val", "test")
_LABEL_TYPE_INFO_100K = {
    "det": ("Detection 2020", "BOX2D"),
    "lane": ("Lane Marking", "POLYLINE2D"),
    "drivable": ("Drivable Area", "POLYGON"),
}
_SEGMENTATIONS_INFO = {
    "sem": ("sem_seg", "masks"),
    "ins": ("ins_seg", "bitmasks"),
    "pan": ("pan_seg", "bitmasks"),
}


def BDD100K(path: str) -> Dataset:
    """`BDD100K <https://bdd-data.berkeley.edu>`_ dataset.

    The file structure should be like::

        <path>
            bdd100k_images_100k/
                images/
                    100k/
                        test
                        train
                        val
                labels/
                    det_20/
                        det_train.json
                        det_val.json
                    lane/
                        polygons/
                            lane_train.json
                            lane_val.json
                    drivable/
                        polygons/
                            drivable_train.json
                            drivable_val.json

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    return _BDD100K_loader(path, "100k")


def BDD100K_10K(path: str) -> Dataset:
    """`BDD100K_10K <https://bdd-data.berkeley.edu>`_ dataset.

    The file structure should be like::

        <path>
            bdd100k_images_10k/
                images/
                    10k/
                        test/
                            cabc30fc-e7726578.jpg
                            ...
                        train/
                            0a0a0b1a-7c39d841.jpg
                            ...
                        val/
                            b1c9c847-3bda4659.jpg
                            ...
                labels/
                    pan_seg/
                        polygons/
                            pan_seg_train.json
                            pan_seg_val.json
                        bitmasks/
                            train/
                                0a0a0b1a-7c39d841.png
                                ...
                            val/
                                b1c9c847-3bda4659.png
                                ...
                    sem_seg/
                        masks/
                            train/
                                0a0a0b1a-7c39d841.png
                                ...
                            val/
                                b1c9c847-3bda4659.png
                                ...
                    ins_seg/
                        bitmasks/
                            train/
                                0a0a0b1a-7c39d841.png
                                ...
                            val/
                                b1c9c847-3bda4659.png
                                ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    return _BDD100K_loader(path, "10k")


def _BDD100K_loader(path: str, dataset_type: str) -> Dataset:
    root_path = os.path.join(
        os.path.abspath(os.path.expanduser(path)), f"bdd100k_images_{dataset_type}"
    )
    dataset = Dataset(DATASET_NAMES[dataset_type])
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), f"catalog_{dataset_type}.json"))
    load_segment = _load_segment_10k if dataset_type == "10k" else _load_segment_100k
    labels_dir = os.path.join(root_path, "labels")
    load_segment(dataset, root_path, labels_dir)

    return dataset


def _load_segment_100k(dataset: Dataset, root_path: str, labels_dir: str) -> None:
    for segment_name in _SEGMENT_NAMES:
        segment = dataset.create_segment(segment_name)
        image_paths = glob(os.path.join(root_path, "images", "100k", segment_name, "*.jpg"))

        print(f"Reading data to segment '{segment_name}'...")
        if segment_name == "test":
            for image_path in image_paths:
                segment.append(Data(image_path))
        else:
            label_contents = _read_label_file_100k(labels_dir, segment_name)
            for image_path in image_paths:
                data = Data(image_path)
                box2d: List[LabeledBox2D] = []
                polygon: List[LabeledPolygon] = []
                polyline2d: List[LabeledPolyline2D] = []
                label = data.label
                label_content = label_contents[os.path.basename(image_path)]
                label.classification = Classification(attributes=label_content["attributes"])
                for label_info in label_content["labels"]:
                    if "box2d" in label_info:
                        _add_box2d_label(label_info, box2d)
                    if "poly2d" in label_info:
                        _add_poly2d_label_100k(label_info, polygon, polyline2d)
                label.box2d = box2d
                label.polygon = polygon
                label.polyline2d = polyline2d
                segment.append(data)
        print(f"Finished reading data to segment '{segment_name}'")


def _load_segment_10k(dataset: Dataset, root_path: str, labels_dir: str) -> None:
    for segment_name in _SEGMENT_NAMES:
        segment = dataset.create_segment(segment_name)
        image_paths = glob(os.path.join(root_path, "images", "10k", segment_name, "*.jpg"))

        print(f"Reading data to segment '{segment_name}'...")
        if segment_name == "test":
            for image_path in image_paths:
                segment.append(Data(image_path))
        else:
            single_channel_mask_dirs: Dict[str, str] = {}
            original_mask_dirs: Dict[str, str] = {}
            for seg_type, dir_names in _SEGMENTATIONS_INFO.items():
                original_mask_dirs[seg_type] = os.path.join(labels_dir, *dir_names, segment_name)
                if seg_type != "sem":
                    single_channel_mask_dir = os.path.join(
                        labels_dir,
                        "single_channel_mask",
                        segment_name,
                        dir_names[0],
                    )
                    single_channel_mask_dirs[seg_type] = single_channel_mask_dir
                    os.makedirs(single_channel_mask_dir, exist_ok=True)

            label_contents = _read_label_file_10k(labels_dir, segment_name)
            for image_path in image_paths:
                segment.append(
                    _get_data_10k(
                        image_path,
                        original_mask_dirs,
                        label_contents[os.path.basename(image_path)],
                        single_channel_mask_dirs,
                    )
                )
            print(f"Finished reading data to segment '{segment_name}'")


def _get_data_10k(
    image_path: str,
    original_mask_paths: Dict[str, str],
    label_content: Dict[str, Any],
    single_channel_mask_paths: Dict[str, str],
) -> Data:
    data = Data(image_path)
    polygon: List[LabeledPolygon] = []
    for label_info in label_content["labels"]:
        if "poly2d" in label_info:
            _add_poly2d_label_10k(label_info, polygon)
    label = data.label
    label.polygon = polygon
    stem = os.path.splitext(os.path.basename(image_path))[0]
    label.semantic_mask = SemanticMask(os.path.join(original_mask_paths["sem"], f"{stem}.png"))
    label.instance_mask = _get_instance_mask(
        stem, original_mask_paths["ins"], single_channel_mask_paths["ins"]
    )
    label.panoptic_mask = _get_panoptic_mask(
        stem, original_mask_paths["pan"], single_channel_mask_paths["pan"]
    )
    return data


def _add_box2d_label(label_info: Dict[str, Any], box2d: List[LabeledBox2D]) -> None:
    box2d_info = label_info["box2d"]
    labeled_box2d = LabeledBox2D(
        box2d_info["x1"],
        box2d_info["y1"],
        box2d_info["x2"],
        box2d_info["y2"],
        category=label_info["category"],
        attributes=label_info["attributes"],
    )
    box2d.append(labeled_box2d)


def _add_poly2d_label_100k(
    label_info: Dict[str, Any], polygon: List[LabeledPolygon], polyline2d: List[LabeledPolyline2D]
) -> None:
    poly2d_info = label_info["poly2d"][0]
    if poly2d_info["closed"]:
        labeled_polygon = LabeledPolygon(
            points=poly2d_info["vertices"],
            category=label_info["category"],
        )
        polygon.append(labeled_polygon)
    else:
        attributes = label_info["attributes"]
        del attributes["laneTypes"]
        labeled_polyline2d = LabeledPolyline2D(
            points=poly2d_info["vertices"],
            category=label_info["category"],
            attributes=attributes,
            beizer_point_types=poly2d_info["types"],
        )
        polyline2d.append(labeled_polyline2d)


def _add_poly2d_label_10k(label_info: Dict[str, Any], polygon: List[LabeledPolygon]) -> None:
    poly2d_info = label_info["poly2d"][0]
    labeled_polygon = LabeledPolygon(
        points=poly2d_info["vertices"],
        category=label_info["category"],
        attributes=label_info.get("attributes", {}),
    )
    polygon.append(labeled_polygon)


def _read_label_file_100k(label_dir: str, segment_name: str) -> Dict[str, Any]:
    source_label_contents = []
    label_filenames = glob(os.path.join(label_dir, "**", f"*_{segment_name}.json"), recursive=True)

    label_prefixes = set(_LABEL_TYPE_INFO_100K)
    for label_filename in label_filenames:
        label_file_basename = os.path.basename(label_filename)
        label_prefix = label_file_basename.replace(f"_{segment_name}.json", "")
        try:
            label_prefixes.remove(label_prefix)
        except KeyError:
            warn_message = f"Invalid label file name '{label_file_basename}'! Ignoring.."
            warn(warn_message)
            continue

        label_description = _LABEL_TYPE_INFO_100K[label_prefix][0]
        print(f"Reading '{label_description}' labels to segment '{segment_name}'...")
        with open(label_filename, "r", encoding="utf-8") as fp:
            source_label_contents.append(json.load(fp))
        print(f"Finished reading '{label_description}' labels to segment '{segment_name}'...")

    for missing_label_prefix in label_prefixes:
        warn_message = (
            f"Missing label file '{missing_label_prefix}_{segment_name}.json'! "
            f"The correspondent '{_LABEL_TYPE_INFO_100K[missing_label_prefix][1]}' "
            f"label will be set to empty!"
        )
        warn(warn_message)

    print(f"Merging '{segment_name}' labels...")
    label_contents = _merge_label(source_label_contents)
    print(f"Finished merging '{segment_name}' labels")
    return label_contents


def _read_label_file_10k(label_dir: str, segment_name: str) -> Dict[str, Any]:
    source_label_contents = []
    label_filename = os.path.join(label_dir, "pan_seg", "polygons", f"pan_seg_{segment_name}.json")
    with open(label_filename, "r", encoding="utf-8") as fp:
        source_label_contents.append(json.load(fp))

    print(f"Merging '{segment_name}' labels...")
    label_contents = _merge_label(source_label_contents)
    print(f"Finished merging '{segment_name}' labels")
    return label_contents


def _merge_label(source_label_contents: List[List[Dict[str, Any]]]) -> Dict[str, Any]:
    label_contents: Dict[str, Any] = {}
    for source_label_content in source_label_contents:
        for image_info in source_label_content:
            image_name = image_info["name"]
            image_label = label_contents.setdefault(image_name, {})
            image_label.setdefault("labels", []).extend(image_info.get("labels", []))
            image_label.setdefault("attributes", {}).update(image_info.get("attributes", {}))

    return label_contents


def _get_instance_mask(stem: str, original_mask_dir: str, mask_dir: str) -> InstanceMask:
    mask_path = os.path.join(mask_dir, f"{stem}.png")
    mask_info = _save_and_get_mask_info(
        os.path.join(original_mask_dir, f"{stem}.png"),
        mask_path,
        os.path.join(mask_dir, f"{stem}.json"),
        "ins",
    )

    ins_mask = InstanceMask(mask_path)
    ins_mask.all_attributes = mask_info["all_attributes"]
    return ins_mask


def _get_panoptic_mask(stem: str, original_mask_dir: str, mask_dir: str) -> PanopticMask:
    mask_path = os.path.join(mask_dir, f"{stem}.png")
    mask_info = _save_and_get_mask_info(
        os.path.join(original_mask_dir, f"{stem}.png"),
        mask_path,
        os.path.join(mask_dir, f"{stem}.json"),
        "pan",
    )

    pan_mask = PanopticMask(mask_path)
    pan_mask.all_category_ids = mask_info["all_category_ids"]
    pan_mask.all_attributes = mask_info["all_attributes"]
    return pan_mask


def _save_and_get_mask_info(
    original_mask_path: str, mask_path: str, mask_info_path: str, seg_type: str
) -> Dict[str, Any]:
    if not os.path.exists(mask_path):
        mask = np.array(Image.open(original_mask_path))
        all_attributes = {}
        if seg_type == "pan":
            all_category_ids = {}
        for category_id, attributes, _, instance_id in np.unique(np.reshape(mask, (-1, 4)), axis=0):
            instance_id = int(instance_id)
            all_attributes[instance_id] = {
                "truncated": bool(attributes & 8),
                "occluded": bool(attributes & 4),
                "crowd": bool(attributes & 2),
                "ignore": bool(attributes & 1),
            }
            if seg_type == "pan":
                all_category_ids[instance_id] = int(category_id)
        mask_info = (
            {"all_attributes": all_attributes, "all_category_ids": all_category_ids}
            if seg_type == "pan"
            else {"all_attributes": all_attributes}
        )
        with open(mask_info_path, "w", encoding="utf-8") as fp:
            json.dump(mask_info, fp)
        Image.fromarray(mask[:, :, -1]).save(mask_path)
    else:
        with open(mask_info_path, "r", encoding="utf-8") as fp:
            mask_info = json.load(
                fp,
                object_hook=lambda info: {
                    int(key) if key.isdigit() else key: value for key, value in info.items()
                },
            )
    return mask_info



#!/usr/bin/env python3
#
# Copytright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the CompCars dataset."""

from tensorbay.opendataset.CompCars.loader import CompCars

__all__ = ["CompCars"]
